import functools
import heapq
from .extensions.merge_methods import get_converter_paths
from .extensions.model_configs import ModelConfig
from .recipe_nodes import RecipeNode, RecipeNodeOrValue
from .graph_finalization import open_graph


def convert(recipe: RecipeNodeOrValue, config: str | ModelConfig | RecipeNode) -> RecipeNodeOrValue:
    """
    Convert a recipe from one model config to another.

    This searches for a chain of registered conversion functions that transform `recipe`’s underlying
    config into the target config, then composes them. For example, you might need to
    convert a LoRA adapter into the base model’s format.

    Args:
        recipe:
            A `RecipeNode` or dictionary representing the input model or partial recipe.
        config (str, ModelConfig or RecipeNode):
            The desired output config, or a recipe node that has the desired config.

    Returns:
        A new recipe node describing the entire conversion.

    Raises:
        TypeError:
            If no conversion path is found.
    """
    if config is None:
        return recipe

    if isinstance(config, RecipeNode):
        with open_graph(config, root_only=True, solve_merge_space=False) as config_graph:
            config = config_graph.finalize(check_mandatory_keys=False, check_extra_keys=False).root.model_config

    tgt_config = config if isinstance(config, str) else config.identifier
    converter_paths = get_converter_paths()

    with open_graph(recipe, root_only=True, solve_merge_space=False) as recipe_graph:
        src_candidates = recipe_graph.root_candidates(model_config_preference=(tgt_config,)).model_config
        for src_config in src_candidates:
            res = create_conversion_recipe(recipe, converter_paths, src_config.identifier, tgt_config)
            if res is not None:
                return res
        raise TypeError(f"Could not convert recipe to config {tgt_config}.")


def create_conversion_recipe(recipe, paths, src_config, tgt_config):
    shortest_path = dijkstra(paths, src_config, tgt_config)
    if shortest_path is None:
        return None
    return functools.reduce(lambda v, mm: mm(v), shortest_path, recipe)


def dijkstra(graph, start, goal):
    """
    graph: Dict[str, List[Tuple[str, any_id]]]
        For each node (str), a list of (neighbor_node, edge_id).
    start: str
    goal: str

    Returns: List of edge IDs (in order) that forms the shortest path from start to goal,
             or None if no path exists.
    """

    distances = {node: float('inf') for node in graph}
    distances[start] = 0
    predecessors = {node: None for node in graph}  # will store the node we came from
    edge_used = {node: None for node in graph}  # will store which edge ID led here
    heap = [(0, start)]

    while heap:
        current_dist, current_node = heapq.heappop(heap)
        if current_dist > distances.get(current_node, float("inf")):
            continue

        if current_node == goal:
            break

        for neighbor, edge_id in graph.get(current_node, ()):
            distance_via_current = current_dist + 1
            if distance_via_current < distances.get(neighbor, float("inf")):
                distances[neighbor] = distance_via_current
                predecessors[neighbor] = current_node
                edge_used[neighbor] = edge_id
                heapq.heappush(heap, (distance_via_current, neighbor))

    if distances.get(goal, float("inf")) == float("inf"):
        return None

    path_ids = []
    node = goal
    while node != start:
        path_ids.append(edge_used[node])
        node = predecessors[node]

    path_ids.reverse()
    return path_ids



import functools
import torch
from typing import Set, Dict, Tuple
from torch import Tensor
from scipy.optimize import linear_sum_assignment
from sd_mecha.extensions.merge_methods import merge_method, Parameter, Return, StateDict
from sd_mecha.extensions import model_configs
from sd_mecha.keys_map import KeyMapBuilder
from sd_mecha.streaming import TensorMetadata
from .perm_graph import PermGraph, PermSlice, Permutation
from .interface import rebasin, randperm


sdxl_config = model_configs.resolve("sdxl-sgm")


def create_graph():
    metadata = sdxl_config.metadata()
    perm_graph = PermGraph(metadata)
    for key, meta in metadata.items():
        num_heads = _num_attention_heads(key)

        if key.startswith("conditioner.embedders.0"):
            clip_l_perms(key, meta, perm_graph, num_heads)

        if key.startswith("conditioner.embedders.1"):
            clip_g_perms(key, meta, perm_graph, num_heads)

        if key.startswith("model.diffusion_model"):
            unet_perms(key, meta, perm_graph, num_heads)

    return perm_graph


def clip_l_perms(key: str, meta: TensorMetadata, perm_graph: PermGraph, num_heads: int):
    if ".q_proj." in key and key.endswith("weight"):  # qk
        head_width = meta.shape[0] // num_heads

        q_key = key
        q_bias_key = q_key.replace(".weight", ".bias")
        k_key = q_key.replace(".q_proj.", ".k_proj.")
        k_bias_key = k_key.replace(".weight", ".bias")

        for start_i in range(0, meta.shape[0], head_width):
            slice = PermSlice(0, start_i)
            perm_graph.add(Permutation(head_width, {
                k: slice for k in (q_key, q_bias_key, k_key, k_bias_key)
            }))

    if ".v_proj." in key and key.endswith("weight"):  # vo
        head_width = meta.shape[0] // num_heads

        v_key = key
        v_bias_key = v_key.replace(".weight", ".bias")
        out_key = v_key.replace(".v_proj.", ".out_proj.")

        for start_i in range(0, meta.shape[0], head_width):
            v_slice = PermSlice(0, start_i)
            perm_graph.add(Permutation(head_width, {
                v_key: v_slice,
                v_bias_key: v_slice,
                out_key: PermSlice(1, start_i),
            }))

    if ".mlp.fc1" in key and key.endswith(".weight"):  # ff
        perm_width = meta.shape[0]

        fc1_key = key
        fc1_bias_key = fc1_key.replace(".weight", ".bias")
        fc2_key = fc1_key.replace(".fc1.", ".fc2.")

        fc1_slice = PermSlice(0)
        fc2_slice = PermSlice(1)
        perm_graph.add(Permutation(perm_width, {
            fc1_key: fc1_slice,
            fc1_bias_key: fc1_slice,
            fc2_key: fc2_slice,
        }))


def clip_g_perms(key: str, meta: TensorMetadata, perm_graph: PermGraph, num_heads: int):
    if ".in_proj_" in key and key.endswith("weight"):  # qkvo
        unfused_size = meta.shape[0] // 3
        head_width = unfused_size // num_heads
        k_offset = unfused_size
        v_offset = k_offset * 2

        in_key = key
        in_bias_key = key.replace("_weight", "_bias")
        out_key = key.replace(".in_proj_weight", ".out_proj.weight")

        for start_i in range(0, unfused_size, head_width):
            # qk
            q_slice = PermSlice(0, start_i)
            k_slice = PermSlice(0, start_i + k_offset)
            perm_graph.add(Permutation(head_width, {
                in_key: (q_slice, k_slice),
                in_bias_key: (q_slice, k_slice),
            }))

            # vo
            v_slice = PermSlice(0, start_i + v_offset)
            out_slice = PermSlice(1, start_i)
            perm_graph.add(Permutation(head_width, {
                in_key: v_slice,
                in_bias_key: v_slice,
                out_key: out_slice,
            }))

    if ".mlp.c_fc." in key and key.endswith(".weight"):  # ff
        perm_width = meta.shape[0]

        fc1_key = key
        fc1_bias_key = key.replace(".weight", ".bias")
        fc2_key = key.replace(".c_fc.", ".c_proj.")

        fc1_slice = PermSlice(0)
        fc2_slice = PermSlice(1)
        perm_graph.add(Permutation(perm_width, {
            fc1_key: fc1_slice,
            fc1_bias_key: fc1_slice,
            fc2_key: fc2_slice,
        }))


def unet_perms(key: str, meta: TensorMetadata, perm_graph: PermGraph, num_heads: int):
    if ".to_q." in key and key.endswith("weight"):  # unet qk
        head_width = meta.shape[0] // num_heads

        q_key = key
        k_key = key.replace(".to_q.", ".to_k.")

        for start_i in range(0, meta.shape[0], head_width):
            slice = PermSlice(0, start_i)
            perm_graph.add(Permutation(head_width, {
                k: slice for k in (q_key, k_key)
            }))

    if ".to_v." in key and key.endswith("weight"):  # unet vo
        head_width = meta.shape[0] // num_heads

        v_key = key
        out_key = key.replace(".to_v.", ".to_out.0.")

        for start_i in range(0, meta.shape[0], head_width):
            v_slice = PermSlice(0, start_i)
            out_slice = PermSlice(1, start_i)
            perm_graph.add(Permutation(head_width, {
                v_key: v_slice,
                out_key: out_slice,
            }))

    if ".ff.net.2." in key and key.endswith(".weight"):  # unet ff
        perm_width = meta.shape[1]

        fc1_key = key.replace(".ff.net.2.", ".ff.net.0.proj.")
        fc1_bias_key = fc1_key.replace(".weight", ".bias")
        fc2_key = key

        fc1_slices = PermSlice(0), PermSlice(0, perm_width)
        fc2_slice = PermSlice(1)
        perm_graph.add(Permutation(perm_width, {
            fc1_key: fc1_slices,
            fc1_bias_key: fc1_slices,
            fc2_key: fc2_slice,
        }))

    if ".time_embed.0." in key and key.endswith(".weight"):  # unet timestep embed
        perm_width = meta.shape[0]

        fc1_key = key
        fc1_bias_key = fc1_key.replace(".weight", ".bias")
        fc2_key = key.replace(".time_embed.0.", ".time_embed.2.")

        fc1_slice = PermSlice(0)
        fc2_slice = PermSlice(1)
        perm_graph.add(Permutation(perm_width, {
            fc1_key: fc1_slice,
            fc1_bias_key: fc1_slice,
            fc2_key: fc2_slice,
        }))

    if ".label_emb.0.0." in key and key.endswith(".weight"):  # unet label embed
        perm_width = meta.shape[0]

        fc1_key = key
        fc1_bias_key = key.replace(".weight", ".bias")
        fc2_key = key.replace(".label_emb.0.0.", ".label_emb.0.2.")

        fc1_slice = PermSlice(0)
        fc2_slice = PermSlice(1)
        perm_graph.add(Permutation(perm_width, {
            fc1_key: fc1_slice,
            fc1_bias_key: fc1_slice,
            fc2_key: fc2_slice,
        }))


def _num_attention_heads(key: str) -> int:
    if key.startswith("conditioner.embedders.0"):
        return 12
    if key.startswith("conditioner.embedders.1"):
        return 20
    return 64


perm_graph = create_graph()


@merge_method(implements=rebasin)
class rebasin_sdxl_sgm:
    @staticmethod
    def map_keys(b: KeyMapBuilder):
        used: Set[str] = set()

        for closure in perm_graph.closures():
            b[closure.keys] = b.keys[closure.keys] @ closure
            used.update(closure.keys)

        for k in b.keys():
            if k not in used:
                b[k] = b.keys[k] @ None

    def __call__(
        self,
        a: Parameter(StateDict[Tensor], model_config=sdxl_config),
        ref: Parameter(StateDict[Tensor], model_config=sdxl_config),
        iters: Parameter(int, model_config=sdxl_config),
        **kwargs,
    ) -> Return(StateDict[Tensor], model_config=sdxl_config):
        out_keys, closure = _get_request(kwargs)

        if closure is None:
            return {k: a[k] for k in out_keys}

        load_a = _make_loader(a)
        load_ref = _make_loader(ref)
        device, dtype = _infer_device_dtype(load_a, closure)

        pid_perm = _init_pid_perm(closure, device)
        _solve_all_components(closure, pid_perm, load_a, load_ref, device, dtype, iters)
        out_all = _apply_inverse_all(closure, pid_perm, load_a)

        return {k: out_all[k] for k in out_keys}


def _infer_device_dtype(load_a, closure):
    rep_key = closure.keys[0]
    t = load_a(rep_key)
    return t.device, t.dtype


def _solve_all_components(
    closure,
    pid_perm: Dict[int, Tensor],
    load_a,
    load_ref,
    device,
    dtype,
    iters: int,
) -> None:
    for comp in closure.components:
        _solve_component(closure, comp, pid_perm, load_a, load_ref, device, dtype, iters)


def _solve_component(
    closure,
    comp: Tuple[int, ...],
    pid_perm: Dict[int, Tensor],
    load_a,
    load_ref,
    device,
    dtype,
    iters: int,
) -> None:
    apps_by_key = _apps_by_key_for_component(closure, comp)

    for _ in range(max(1, iters)):
        changed = False

        for pid in comp:
            new_p = _update_one_pid(
                closure, pid, pid_perm, apps_by_key, load_a, load_ref, device, dtype
            )
            if not torch.equal(new_p, pid_perm[pid]):
                pid_perm[pid] = new_p
                changed = True

        if not changed:
            break


def _apps_by_key_for_component(closure, comp: Tuple[int, ...]):
    apps_by_key: Dict[str, Tuple[Tuple[int, int, int], ...]] = {}
    comp_set = set(comp)

    for k in closure.keys:
        apps = closure.key_to_apps.get(k)
        if not apps:
            continue
        filt = tuple(app for app in apps if app[0] in comp_set)
        if filt:
            apps_by_key[k] = filt

    return apps_by_key


def _update_one_pid(
    closure,
    pid: int,
    pid_perm: Dict[int, Tensor],
    apps_by_key: Dict[str, Tuple[Tuple[int, int, int], ...]],
    load_a,
    load_ref,
    device,
    dtype,
) -> Tensor:
    w = closure.pid_to_width[pid]
    nodes = closure.pid_to_nodes[pid]

    cost = torch.zeros((w, w), device=device, dtype=dtype)

    for n in nodes:
        k = n.key
        axis_cur = _norm_axis(n.axis, load_a(k).dim())
        offset = n.offset

        tensor_a = _a_vec(load_a(k), axis_cur, offset, w)
        tensor_b = _b_vec(load_ref(k), apps_by_key.get(k, ()), axis_cur, offset, w, pid_perm, closure)
        cost.addmm_(tensor_a, tensor_b.T)

    return _solve_lap_max(cost)


def _a_vec(tensor_a: Tensor, axis_cur: int, offset: int, w: int) -> Tensor:
    a_window = _slice_window(tensor_a, axis_cur, offset, w)
    return _front_flat(a_window, axis_cur)


def _b_vec(
    tensor_ref: Tensor,
    apps: Tuple[Tuple[int, int, int], ...],
    axis_cur: int,
    offset: int,
    width: int,
    pid_perm: Dict[int, Tensor],
    closure,
) -> Tensor:
    b_ref_other = _apply_other_axis_perms(tensor_ref, apps, axis_cur, pid_perm, closure)
    b_window = _slice_window(b_ref_other, axis_cur, offset, width)
    return _front_flat(b_window, axis_cur)


def _apply_other_axis_perms(
    tensor_ref: Tensor,
    apps: Tuple[Tuple[int, int, int], ...],
    axis_cur: int,
    pid_perm: Dict[int, Tensor],
    closure,
) -> Tensor:
    # Clone only if we actually apply something
    out = tensor_ref
    cloned = False

    for pid2, ax2, off2 in apps:
        ax2n = _norm_axis(ax2, tensor_ref.dim())
        if ax2n == axis_cur:
            continue

        if not cloned:
            out = tensor_ref.clone()
            cloned = True

        _apply_window_perm_inplace(
            out,
            axis=ax2n,
            offset=off2,
            width=closure.pid_to_width[pid2],
            perm=pid_perm[pid2],
        )

    return out


def _apply_inverse_all(closure, pid_perm: Dict[int, Tensor], load_a) -> Dict[str, Tensor]:
    out_all: Dict[str, Tensor] = {}

    for k in closure.keys:
        tensor_a = load_a(k)
        apps = closure.key_to_apps.get(k)
        if not apps:
            out_all[k] = tensor_a
            continue

        out_all[k] = _apply_inverse_one(tensor_a, apps, closure, pid_perm)

    return out_all


def _apply_inverse_one(
    tensor_a: Tensor,
    apps: Tuple[Tuple[int, int, int], ...],
    closure,
    pid_perm: Dict[int, Tensor],
) -> Tensor:
    out = tensor_a.clone()
    for pid, ax, off in apps:
        w = closure.pid_to_width[pid]
        inv = _invert_perm(pid_perm[pid])
        _apply_window_perm_inplace(out, axis=ax, offset=off, width=w, perm=inv)
    return out


@merge_method(implements=randperm)
class randperm_sdxl_sgm:
    @staticmethod
    def map_keys(b: KeyMapBuilder):
        used: Set[str] = set()

        for closure in perm_graph.closures():
            b[closure.keys] = b.keys[closure.keys] @ closure
            used.update(closure.keys)

        for k in b.keys():
            if k not in used:
                b[k] = b.keys[k] @ None

    def __call__(
        self,
        a: Parameter(StateDict[Tensor], model_config=sdxl_config),
        seed: Parameter(int, model_config=sdxl_config) = None,
        **kwargs,
    ) -> Return(StateDict[Tensor], model_config=sdxl_config):
        out_keys, closure = _get_request(kwargs)

        if closure is None:
            return {k: a[k] for k in out_keys}

        load_a = _make_loader(a)
        device = load_a(closure.keys[0]).device

        pid_perm = _init_rand_perm(closure, seed, device)
        out_all = _apply_all_perms(closure, pid_perm, load_a)

        return {k: out_all[k] for k in out_keys}


def _get_request(kwargs) -> Tuple[Tuple[str, ...], "PermClosure | None"]:
    rel = kwargs["key_relation"]
    return rel.outputs, rel.meta


def _make_loader(a):
    return functools.partial(_load, memo={}, sd=a)


def _init_pid_perm(closure, device) -> Dict[int, Tensor]:
    pid_perm: Dict[int, Tensor] = {}

    for pid in closure.perms:
        w = closure.pid_to_width[pid]
        pid_perm[pid] = torch.arange(w, dtype=torch.long, device=device)

    return pid_perm


def _init_rand_perm(closure, seed: int | None, device) -> Dict[int, Tensor]:
    pid_perm: Dict[int, Tensor] = {}

    for pid in closure.perms:
        w = closure.pid_to_width[pid]
        gen = _make_generator(seed, pid, device)
        pid_perm[pid] = torch.randperm(w, generator=gen, device=device)

    return pid_perm


def _make_generator(seed: int | None, pid: int, device):
    if seed is None:
        return None
    g = torch.Generator(device=device)
    g.manual_seed(int(seed) + int(pid))
    return g


def _apply_all_perms(closure, pid_perm: Dict[int, Tensor], load_a) -> Dict[str, Tensor]:
    out_all: Dict[str, Tensor] = {}

    for k in closure.keys:
        t = load_a(k)
        apps = closure.key_to_apps.get(k)
        if not apps:
            out_all[k] = t
            continue

        out_all[k] = _apply_one_key(t, apps, closure, pid_perm)

    return out_all


def _apply_one_key(
    t: Tensor,
    apps: Tuple[Tuple[int, int, int], ...],
    closure,
    pid_perm: Dict[int, Tensor],
) -> Tensor:
    out = t.clone()
    for pid, ax, off in apps:
        w = closure.pid_to_width[pid]
        _apply_window_perm_inplace(
            out,
            axis=ax,
            offset=off,
            width=w,
            perm=pid_perm[pid],
        )
    return out


def _load(k: str, memo, sd) -> Tensor:
    t = memo.get(k)
    if t is None:
        t = sd[k]
        memo[k] = t
    return t


def _front_flat(t: Tensor, axis: int) -> Tensor:
    t = torch.movedim(t, axis, 0)
    return t.reshape(t.shape[0], -1).contiguous()


def _solve_lap_max(sim: Tensor) -> Tensor:
    """Return perm p s.t. sum_i sim[i, p[i]] is maximized. sim on CPU."""
    if sim.dtype.itemsize < 2:
        sim = sim.float()
    r, c = linear_sum_assignment((-sim).numpy(force=True))
    return torch.as_tensor(c, device=sim.device, dtype=torch.long)


def _apply_window_perm_inplace(
    t: Tensor,
    *,
    axis: int,
    offset: int,
    width: int,
    perm: Tensor,
) -> None:
    ax = _norm_axis(axis, t.dim())
    win = _slice_window(t, ax, offset, width)
    winp = torch.index_select(win, ax, perm)
    idx = [slice(None)] * t.dim()
    idx[ax] = slice(offset, offset + width)
    t[tuple(idx)] = winp


def _slice_window(t: Tensor, axis: int, offset: int, width: int) -> Tensor:
    ax = _norm_axis(axis, t.dim())
    idx = [slice(None)] * t.dim()
    idx[ax] = slice(offset, offset + width)
    return t[tuple(idx)]


def _norm_axis(axis: int, rank: int) -> int:
    if axis < 0:
        axis += rank
    if axis < 0 or axis >= rank:
        raise ValueError(f"axis {axis} out of range for rank {rank}")
    return axis


def _invert_perm(p: Tensor) -> Tensor:
    inv = torch.empty_like(p)
    inv[p] = torch.arange(p.numel(), device=p.device, dtype=p.dtype)
    return inv



from torch import Tensor
from sd_mecha.extensions.merge_methods import merge_method, Parameter, Return, StateDict


@merge_method(is_interface=True)
def balance_attention(
    a: Parameter(StateDict[Tensor]),
    **kwargs,
) -> Return(Tensor):
    ...



def _load_builtin_extensions():
    import sd_mecha.extensions.builtin.model_configs
    import sd_mecha.extensions.builtin.lycoris


_load_builtin_extensions()


from .recipe_nodes import RecipeNodeOrValue, RecipeNode
from .merging import merge
from .graph_finalization import open_graph
from .serialization import serialize, deserialize, deserialize_path
from .streaming import StateDictKeyError
from .extensions.merge_methods import merge_method, value_to_node, Parameter, Return, StateDict
from .extensions import model_dirs
from .conversion import convert
from sd_mecha.extensions.builtin.merge_methods import (
    clamp,
    add_cosine_a,
    add_cosine_b,
    crossover,
    exchange_ema,
    weighted_sum,
    n_average,
    slerp,
    subtract,
    perpendicular_component,
    train_difference_mask,
    add_opposite_mask,
    add_strict_opposite_mask,
    geometric_sum,
    multiply_quotient,
    fallback,
    cast,
    get_dtype,
    get_device,
    pick_component,
    omit_component,
    cast_dtype_map,
    cast_dtype_map_reversed,
    stack,
    omit_non_finite,
    rebasin,
    randperm,
    balance_attention,
    top_k_tensor_sum,
    truncate_rank,
    ties_sum_with_dropout,
    ties_sum,
    ties_sum_extended,
    model_stock,
    geometric_median,
    add_difference,
    add_perpendicular,
    add_difference_ties,
    add_difference_ties_extended,
    copy_region,
    tensor_sum,
    rotate,
    dropout,
    add_ties_with_dare,
    n_model_stock,
)
from .helpers import model, literal, Defaults, set_log_level, skip_key
from . import recipe_nodes, extensions, merge_context



from typing import Iterable, Tuple


def convert_clip_g_key(sgm_key: str) -> Tuple[Iterable[str], bool]:
    transpose = False
    if sgm_key.endswith("text_projection"):
        kohya_key = "te2.text_projection.weight"
    else:
        kohya_key = sgm_key.replace("conditioner.embedders.1.model.", "te2.text_model.")
        kohya_key = kohya_key.replace(".token_embedding.", ".embeddings.token_embedding.")
        kohya_key = kohya_key.replace(".positional_embedding", ".embeddings.position_embedding.weight")
        kohya_key = kohya_key.replace(".transformer.resblocks.", ".encoder.layers.")
        kohya_key = kohya_key.replace(".attn.", ".self_attn.")
        kohya_key = kohya_key.replace(".mlp.c_fc.", ".mlp.fc1.")
        kohya_key = kohya_key.replace(".mlp.c_proj.", ".mlp.fc2.")
        kohya_key = kohya_key.replace(".ln_final.", ".final_layer_norm.")
        kohya_key = kohya_key.replace(".ln_", ".layer_norm")

    if kohya_key.endswith((".in_proj_weight", ".in_proj_bias")):
        is_bias = kohya_key.endswith("bias")
        partial_key = kohya_key.replace(".in_proj_weight", "").replace(".in_proj_bias", "")
        res = tuple((
            f"{partial_key}.{k}_proj.{'bias' if is_bias else 'weight'}"
            for k in ("q", "k", "v")
        ))
    else:
        res = (kohya_key,)

    if sgm_key.endswith("text_projection"):
        transpose = True

    return res, transpose



import torch


def balance_head_energy(q: torch.Tensor, k: torch.Tensor, eps: float = 1e-6):
    """
    Wq, Wk: [H, 64, d] (NO bias column here)
    Returns Wq', Wk' with identical logits: (Wq'x)^T (Wk'x) == (Wq x)^T (Wk x)
    """
    a = q @ q.mH
    b = k @ k.mH

    a.diagonal()[:] += eps
    b.diagonal()[:] += eps

    v_a, vs_a = torch.linalg.eigh(a)
    v_a = v_a.clamp_min(eps)
    v_a_half = v_a.pow(0.5)
    v_a_half_neg = v_a.pow(-0.5)
    a_half = (vs_a * v_a_half.unsqueeze(-2)) @ vs_a.mH
    a_half_neg = (vs_a * v_a_half_neg.unsqueeze(-2)) @ vs_a.mH

    c = a_half @ b @ a_half
    v_c, vs_c = torch.linalg.eigh(c)
    v_c_half = v_c.clamp_min(eps).pow(0.5)
    c_half = (vs_c * v_c_half.unsqueeze(-2)) @ vs_c.mT

    s = a_half_neg @ c_half @ a_half_neg
    q = s @ q
    k = torch.linalg.solve(s, k)
    return q, k


def bundle_weight_bias(w, b):
    b = b.unsqueeze(-1)
    wb = torch.cat([w, b], dim=-1)
    return wb


def split_weight_bias(wb):
    b = wb[..., -1]
    w = wb[..., :-1]
    return w, b



import torch
from torch import Tensor
from typing import TypeVar
from sd_mecha.extensions.merge_methods import merge_method, StateDict, Parameter, Return
from sd_mecha.streaming import StateDictKeyError, NonFiniteStateDictKeyError


T = TypeVar("T")


@merge_method(reuse_outputs=False)
def fallback(
    a: Parameter(StateDict[T]),
    default: Parameter(StateDict[T]),
    **kwargs,
) -> Return(T):
    key = kwargs["key"]
    try:
        return a[key]
    except StateDictKeyError:
        return default[key]


@merge_method(reuse_outputs=False)
def cast(
    a: Parameter(Tensor),
    device: Parameter(str) = None,
    dtype: Parameter(str) = None,
) -> Return(Tensor):
    to_kwargs = {}
    if device is not None:
        to_kwargs["device"] = device

    if dtype is not None:
        if dtype not in cast_dtype_map:
            raise ValueError(f"Unknown dtype {dtype}. Possible values are None, {', '.join(cast_dtype_map.keys())}")
        to_kwargs["dtype"] = cast_dtype_map[dtype]

    return a.to(**to_kwargs)


cast_dtype_map = {
    "float64": torch.float64,
    "int64": torch.int64,
    "float32": torch.float32,
    "int32": torch.int32,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
    "int16": torch.int16,
    "int8": torch.int8,
    "bool": torch.bool,
}
for dtype_str in ("uint8", "uint16", "uint32", "uint64", "float8_e4m3fn", "float8_e4m3fnuz", "float8_e5m2", "float8_e5m2fnuz", "float8_e8m0fnu"):
    if hasattr(torch, dtype_str):
        cast_dtype_map[dtype_str] = getattr(torch, dtype_str)
cast_dtype_map_reversed = {v: k for k, v in cast_dtype_map.items()}


@merge_method(reuse_outputs=False)
def get_dtype(
    a: Parameter(Tensor),
) -> Return(str, "param"):
    return cast_dtype_map_reversed[a.dtype]


@merge_method(reuse_outputs=False)
def get_device(
    a: Parameter(Tensor),
) -> Return(str, "param"):
    return str(a.device)


@merge_method(reuse_outputs=False)
def pick_component(
    a: Parameter(StateDict[T]),
    component: Parameter(str, "param"),
    **kwargs,
) -> Return(T):
    if component not in a.model_config.components():
        raise ValueError(
            f'Component "{component}" does not exist in config "{a.model_config.identifier}". '
            f"Valid components: {tuple(a.model_config.components())}"
        )

    key = kwargs["key"]
    if key in a.model_config.components()[component].keys():
        return a[key]
    else:
        raise StateDictKeyError(key)


@merge_method(reuse_outputs=False)
def omit_component(
    a: Parameter(StateDict[T]),
    component: Parameter(str, "param"),
    **kwargs,
) -> Return(T):
    if component not in a.model_config.components():
        raise ValueError(
            f'Component "{component}" does not exist in config "{a.model_config.identifier}". '
            f"Valid components: {tuple(a.model_config.components())}"
        )

    key = kwargs["key"]
    if key in a.model_config.components()[component].keys():
        raise StateDictKeyError(key)
    else:
        return a[key]


@merge_method(reuse_outputs=False)
def stack(
    *values: Parameter(Tensor),
) -> Return(Tensor):
    return torch.stack(values)


@merge_method(reuse_outputs=False)
def omit_non_finite(
    a: Parameter(StateDict[Tensor]),
    omit_mandatory: Parameter(bool) = False,
    omit_optional: Parameter(bool) = True,
    **kwargs,
) -> Return(Tensor):
    key = kwargs["key"]
    v = a[key]
    optional = a.model_config.keys()[key].optional
    check_finite = omit_mandatory and not optional or omit_optional and optional
    if check_finite and not v.isfinite().all():
        raise NonFiniteStateDictKeyError(key)
    else:
        return v



from typing import TypeVar
from sd_mecha.extensions.merge_methods import merge_method, Parameter, Return, StateDict


T = TypeVar("T")


@merge_method(is_conversion=True, reuse_outputs=False)
class convert_singleton:
    @staticmethod
    def map_keys(b):
        b[...] = b.keys["key"]

    def __call__(
        self,
        singleton: Parameter(StateDict[T], model_config="singleton-mecha"),
    ) -> Return(T):
        return singleton["key"]



import re
from typing import TypeVar
from sd_mecha.extensions.merge_methods import merge_method, StateDict, Parameter, Return
from sd_mecha.extensions import model_configs


sdxl_sgm_config = model_configs.resolve("sdxl-sgm")


re_inp = re.compile(r"\.input_blocks\.(\d+)\.")  # 12
re_mid = re.compile(r"\.middle_block\.(\d+)\.")  # 1
re_out = re.compile(r"\.output_blocks\.(\d+)\.")  # 12


T = TypeVar("T")


# srcs, in order of priority when disagreements occur:
# - https://github.com/hako-mikan/sd-webui-supermerger/blob/f14b3e5d0be9c510d199cca502c4148160f901bb/scripts/mergers/mergers.py#L1376
# - https://github.com/s1dlx/meh/blob/04af2c8d63744fb6c02d35d328a2c84380cca444/sd_meh/merge.py#L360
# - https://github.com/vladmandic/automatic/blob/e22d0789bddd3894364b0d59a4c9b3e456e89079/modules/merging/merge_utils.py#L64
@merge_method(is_conversion=True, reuse_outputs=False)
class convert_sdxl_blocks_to_sgm:
    @staticmethod
    def map_keys(b):
        for output_key in sdxl_sgm_config.keys():
            input_key = "BASE"
            if output_key.startswith("model.diffusion_model."):
                input_key = "OUT08"
                if ".time_embed" in output_key or ".label_emb" in output_key:
                    input_key = "BASE"  # before input blocks
                elif ".out." in output_key:
                    input_key = "OUT08"  # after output blocks
                elif m := re_inp.search(output_key):
                    input_key = f"IN{int(m.groups(1)[0]):02}"
                elif re_mid.search(output_key):
                    input_key = "M00"
                elif m := re_out.search(output_key):
                    input_key = f"OUT{int(m.groups(1)[0]):02}"
            elif output_key.startswith("first_stage_model."):
                input_key = "VAE"

            b[output_key] = b.keys[input_key]

    def __call__(
        self,
        blocks: Parameter(StateDict[T], model_config="sdxl-supermerger_blocks"),
        **kwargs,
    ) -> Return(T, model_config="sdxl-sgm"):
        relation = kwargs["key_relation"]
        block_key = relation.inputs["blocks"][0]
        return blocks[block_key]



import re
import torch
from collections import defaultdict
from torch import Tensor
from typing import Optional, Tuple
from sd_mecha.extensions.merge_methods import merge_method, Parameter, Return, StateDict
from sd_mecha.keys_map import KeyMapBuilder
from .algorithms import balance_head_energy, bundle_weight_bias, split_weight_bias
from .interface import balance_attention


def implement_merge_methods(config, base, extract_qkvo, return_qkvo):
    @merge_method(implements=balance_attention)
    class balance_attention_sdxl_sgm(base):
        def __call__(
            self,
            a: Parameter(StateDict[Tensor], model_config=config),
            **kwargs,
        ) -> Return(Tensor, model_config=config):
            keys = kwargs["key_relation"].outputs

            qkvo = extract_qkvo(a, keys)
            if qkvo is None:
                return a[keys[0]]

            q, k, v, o = qkvo
            q, k = balance_head_energy(q, k)
            v, o = balance_head_energy(v, o)
            return return_qkvo(q, k, v, o, keys)


class SdxlSgmAttentionBase:
    clip_l_re = re.compile(r"layers\.(\d+)\.self_attn\.(.+)")
    clip_g_re = re.compile(r"resblocks\.(\d+)\.attn\.(.+)")
    vae_re = re.compile(r"\.([a-z]+)\.mid\.attn_1\.(.+)")
    unet_re = re.compile(r"\.([a-z]+)_blocks?\.(\d+\.)(?:\d+\.)?transformer_blocks\.(\d+)\.attn([12])\.(.+)")

    @classmethod
    def map_keys(cls, b: KeyMapBuilder):
        attention_keys = defaultdict(list)
        for key in b.keys():
            if (match := cls.clip_l_re.search(key)) and match.group(2) != "out_proj.bias":
                layer_id = ("clip_l", match.group(1))
                attention_keys[layer_id].append(key)
            elif (match := cls.clip_g_re.search(key)) and match.group(2) != "out_proj.bias":
                layer_id = ("clip_g", match.group(1))
                attention_keys[layer_id].append(key)
            elif (match := cls.vae_re.search(key)) and match.group(2) != "proj_out.bias":
                layer_id = ("vae", *match.group(1, 2))
                attention_keys[layer_id].append(key)
            elif (match := cls.unet_re.search(key)) and match.group(5) != "to_out.0.bias":
                layer_id = ("unet", *match.group(1, 2, 3, 4))
                attention_keys[layer_id].append(key)
            else:
                b[key] = b.keys[key]

        for group in attention_keys.values():
            b[tuple(group)] = b.keys[tuple(group)]


def extract_qkvo_sdxl_sgm(
    sd: StateDict[Tensor],
    keys: Tuple[str, ...],
) -> Optional[Tuple[Tensor, Tensor, Tensor, Tensor]]:
    if len(keys) == 1:
        return None

    if is_clip_l(keys[0]) or is_vae(keys[0]):
        (
            key_k_bias, key_k,
            key_o,
            key_q_bias, key_q,
            key_v_bias, key_v,
        ) = keys

        k_bias, k = sd[key_k_bias], sd[key_k]
        o = sd[key_o]
        q_bias, q = sd[key_q_bias], sd[key_q]
        v_bias, v = sd[key_v_bias], sd[key_v]

        k = bundle_weight_bias(k, k_bias).unflatten(0, (-1, 64))
        q = bundle_weight_bias(q, q_bias).unflatten(0, (-1, 64))
        v = bundle_weight_bias(v, v_bias).unflatten(0, (-1, 64))
        o = bundle_weight_bias(o.mT, torch.zeros_like(o.mT[..., -1])).unflatten(0, (-1, 64))
    elif is_clip_g(keys[0]):
        (
            key_bias, key_weight,
            key_o,
        ) = keys

        bias = sd[key_bias]
        weight = sd[key_weight]
        o = sd[key_o]

        dim = weight.shape[0] // 3

        q = bundle_weight_bias(weight[:dim], bias[:dim]).unflatten(0, (-1, 64))
        k = bundle_weight_bias(weight[dim:dim*2], bias[dim:dim*2]).unflatten(0, (-1, 64))
        v = bundle_weight_bias(weight[dim*2:], bias[dim*2:]).unflatten(0, (-1, 64))
        o = bundle_weight_bias(o.mT, torch.zeros_like(o.mT[..., -1])).unflatten(0, (-1, 64))
    elif is_unet(keys[0]):
        key_k, key_o, key_q, key_v = keys

        k = sd[key_k].unflatten(0, (-1, 64))
        o = sd[key_o].mT.unflatten(0, (-1, 64))
        q = sd[key_q].unflatten(0, (-1, 64))
        v = sd[key_v].unflatten(0, (-1, 64))
    else:
        raise KeyError(keys)

    return q, k, v, o


def return_qkvo_sdxl_sgm(q, k, v, o, keys):
    if is_clip_l(keys[0]) or is_vae(keys[0]):
        (
            key_k_bias, key_k,
            key_o,
            key_q_bias, key_q,
            key_v_bias, key_v,
        ) = keys

        k, k_bias = split_weight_bias(k.flatten(end_dim=1))
        o = split_weight_bias(o.flatten(end_dim=1))[0].mT
        q, q_bias = split_weight_bias(q.flatten(end_dim=1))
        v, v_bias = split_weight_bias(v.flatten(end_dim=1))

        result = {
            key_k_bias: k_bias, key_k: k,
            key_o: o,
            key_q_bias: q_bias, key_q: q,
            key_v_bias: v_bias, key_v: v,
        }
    elif is_clip_g(keys[0]):
        (
            key_bias, key_weight,
            key_o,
        ) = keys

        k, k_bias = split_weight_bias(k.flatten(end_dim=1))
        o = split_weight_bias(o.flatten(end_dim=1))[0].mT
        q, q_bias = split_weight_bias(q.flatten(end_dim=1))
        v, v_bias = split_weight_bias(v.flatten(end_dim=1))

        weight = torch.cat((q, k, v), dim=0)
        bias = torch.cat((q_bias, k_bias, v_bias), dim=0)

        result = {
            key_bias: bias, key_weight: weight,
            key_o: o,
        }
    elif is_unet(keys[0]):
        (
            key_k,
            key_o,
            key_q,
            key_v,
        ) = keys

        k = k.flatten(end_dim=1)
        o = o.flatten(end_dim=1).mT
        q = q.flatten(end_dim=1)
        v = v.flatten(end_dim=1)

        result = {
            key_k: k,
            key_o: o,
            key_q: q,
            key_v: v,
        }
    else:
        raise KeyError(keys)

    return result


implement_merge_methods("sdxl-sgm", SdxlSgmAttentionBase, extract_qkvo_sdxl_sgm, return_qkvo_sdxl_sgm)


def is_clip_l(k):
    return k.startswith("conditioner.embedders.0")


def is_clip_g(k):
    return k.startswith("conditioner.embedders.1")


def is_vae(k):
    return k.startswith("first_stage_model")


def is_unet(k):
    return k.startswith("model.diffusion_model")



import abc
import dataclasses
import torch
from typing import Iterable, Mapping, Dict
from sd_mecha.extensions import model_configs
from .merge_methods.kronecker import kron_dims_from_ratio
from sd_mecha.extensions.merge_methods import merge_method, StateDict, Parameter, Return
from sd_mecha.extensions.model_configs import StateDictKey, ModelConfig, ModelConfigImpl, LazyModelConfigBase, KeyMetadata
from sd_mecha.streaming import StateDictKeyError
from sd_mecha.keys_map import KeyMapBuilder


def _register_all_lycoris_configs():
    for lyco_id, lyco_prefix in (
        ("lycoris", "lycoris"),
        ("kohya", "lora"),
    ):
        @merge_method(identifier=f"extract_{lyco_id}_lora", is_interface=True)
        def extract_lora(
            base: Parameter(StateDict[torch.Tensor], "delta"),
            rank: Parameter(StateDict[int]) = 8,
        ) -> Return(StateDict[torch.Tensor], "weight"):
            ...

        @merge_method(identifier=f"extract_{lyco_id}_lokr", is_interface=True)
        def extract_lokr(
            base: Parameter(StateDict[torch.Tensor], "delta"),
            kronecker_ratio: Parameter(StateDict[float]) = 0.5,
        ) -> Return(StateDict[torch.Tensor], "weight"):
            ...

        lyco_interfaces = {
            "lora": extract_lora,
            "lokr": extract_lokr,
        }

        for base_config_id in (
            "sdxl-kohya",
            "sdxl-kohya_but_diffusers",
            "sd1-kohya",
        ):
            base_config = model_configs.resolve(base_config_id)
            lyco_config = LycorisModelConfig(base_config, lyco_id, lyco_prefix, list(lycoris_algorithms))
            model_configs.register_aux(lyco_config)
            define_conversions(lyco_config, lyco_interfaces)


def define_conversions(lyco_config, lyco_interfaces):
    lyco_config_id = lyco_config.identifier
    base_config = lyco_config.base_config
    base_config_id = base_config.identifier

    @merge_method(identifier=f"convert_'{lyco_config_id}'_to_base", is_conversion=True)
    class LycorisToBase:
        @staticmethod
        def map_keys(b: KeyMapBuilder):
            for base_key in base_config.keys():
                input_keys = tuple(lyco_config.to_lycoris_keys(base_key))
                if not input_keys:
                    continue
                b[base_key] = b.keys[input_keys]

        def __call__(
            self,
            lora: Parameter(StateDict[torch.Tensor], "weight", lyco_config_id),
            **kwargs,
        ) -> Return(torch.Tensor, "delta", base_config_id):
            (base_key,), input_keys = kwargs["key_relation"]
            target_shape = base_config.keys()[base_key].shape

            key_prefix = input_keys["lora"][0].split(".")[0]
            sd_helper = StateDictKeyHelper(lora, key_prefix)
            for compose_fn in (compose_lora, compose_lokr):
                try:
                    return compose_fn(sd_helper, target_shape)
                except StateDictKeyError:
                    pass

            raise StateDictKeyError(base_key)

    class BaseToLycoris(abc.ABC):
        @classmethod
        def map_keys(cls, b):
            for base_key in base_config.keys():
                if output_keys := cls.get_output_keys(base_key):
                    b[output_keys] = b.keys[base_key]

        @staticmethod
        @abc.abstractmethod
        def get_output_keys(base_key: str):
            ...

    @merge_method(identifier=f"extract_lora_'{lyco_config_id}'", implements=lyco_interfaces["lora"])
    class BaseToLora(BaseToLycoris):
        @staticmethod
        def get_output_keys(base_key: str):
            keys = lyco_config.to_lycoris_keys(base_key, ("lora",))
            if not keys:
                return ()
            up, _, down, alpha = keys
            return up, down, alpha

        def __call__(
            self,
            base: Parameter(StateDict[torch.Tensor], "delta", base_config_id),
            rank: Parameter(StateDict[int], model_config=base_config_id),
            **kwargs,
        ) -> Return(StateDict[torch.Tensor], "weight", lyco_config_id):
            (up_key, down_key, alpha_key), input_keys = kwargs["key_relation"]
            base_key = input_keys["base"][0]

            base_value = base[base_key]
            rank_value = rank[base_key]
            original_shape = base_value.shape
            shape_vh = torch.Size((rank_value, *original_shape[1:]))
            shape_2d = torch.Size((original_shape[0], original_shape[1:].numel()))

            svd_driver = "gesvd" if base_value.is_cuda else None
            u, s, vh = torch.linalg.svd(base_value.reshape(shape_2d), full_matrices=False, driver=svd_driver)
            s = s[..., :rank_value].sqrt()
            u = u[..., :rank_value] * s.unsqueeze(-2)
            vh = s.unsqueeze(-1) * vh[..., :rank_value, :]

            return {
                up_key: u,
                down_key: vh.reshape(shape_vh),
                alpha_key: torch.tensor(rank_value, device=base_value.device, dtype=base_value.dtype),
            }

    @merge_method(identifier=f"extract_lokr_'{lyco_config_id}'", implements=lyco_interfaces["lokr"])
    class BaseToLokr(BaseToLycoris):
        @staticmethod
        def get_output_keys(base_key: str):
            keys = lyco_config.to_lycoris_keys(base_key, ("lokr",))
            if not keys:
                return ()
            w1, _, _, w2, _, _, _, _ = keys
            return w1, w2

        def __call__(
            self,
            base: Parameter(StateDict[torch.Tensor], "delta", base_config_id),
            kronecker_ratio: Parameter(StateDict[float], model_config=base_config_id),
            **kwargs,
        ) -> Return(StateDict[torch.Tensor], "weight", lyco_config_id):
            (w1_key, w2_key), input_keys = kwargs["key_relation"]
            base_key = input_keys["base"][0]

            base_value = base[base_key]
            kronecker_ratio_value = kronecker_ratio[base_key]
            shape_original = base_value.shape
            m1, m2, n1, n2 = kron_dims_from_ratio(shape_original, kronecker_ratio_value)
            shape_w1 = torch.Size((m1, n1))
            shape_w2 = torch.Size((m2, n2, *shape_original[2:]))
            p2 = shape_original[2:].numel()

            value_2d = base_value.reshape(m1, m2, n1, n2*p2).permute(0, 2, 1, 3).reshape(shape_w1.numel(), shape_w2.numel())

            svd_driver = "gesvd" if base_value.is_cuda else None
            u, s, vh = torch.linalg.svd(value_2d, full_matrices=False, driver=svd_driver)
            s = s[..., 0].sqrt()
            u = u[..., 0] * s
            vh = s * vh[..., 0, :]

            return {
                w1_key: u.reshape(shape_w1),
                w2_key: vh.reshape(shape_w2),
            }


class StateDictKeyHelper:
    def __init__(self, state_dict: Mapping[str, torch.Tensor], key_prefix):
        self.state_dict = state_dict
        self.key_prefix = key_prefix

    def get_tensor(self, name, raise_on_missing=True):
        try:
            return self.state_dict[f"{self.key_prefix}.{name}"]
        except StateDictKeyError as e:
            if raise_on_missing:
                raise
            else:
                return None


def compose_lora(state_dict: StateDictKeyHelper, target_shape: torch.Size):
    # fetching these 4 keys in lexicographic order is very important
    # any other order would raise the number of cache misses in the input safetensors when streaming keys
    #   which in turn would slow down merging significantly
    alpha = state_dict.get_tensor("alpha")
    down_weight = state_dict.get_tensor("lora_down.weight")
    mid_weight = state_dict.get_tensor("lora_mid.weight", raise_on_missing=False)
    up_weight = state_dict.get_tensor("lora_up.weight")
    dim = down_weight.size(0)

    if mid_weight is not None:
        wa = up_weight.view(up_weight.size(0), -1).transpose(0, 1)
        wb = down_weight.view(down_weight.size(0), -1)
        delta = rebuild_tucker(mid_weight, wa, wb)
    else:
        delta = up_weight.view(up_weight.size(0), -1) @ down_weight.view(down_weight.size(0), -1)

    delta = delta * (alpha / dim)
    delta = delta.reshape(target_shape)
    return delta


def compose_lokr(state_dict: StateDictKeyHelper, target_shape: torch.Size):
    lora_dim = None

    w1 = state_dict.get_tensor("lokr_w1", raise_on_missing=False)
    if w1 is None:
        w1a = state_dict.get_tensor("lokr_w1_a")
        w1b = state_dict.get_tensor("lokr_w1_b")
        w1 = w1a @ w1b
        lora_dim = w1b.shape[0]

    w2 = state_dict.get_tensor("lokr_w2", raise_on_missing=False)
    if w2 is None:
        w2a = state_dict.get_tensor("lokr_w2_a")
        w2b = state_dict.get_tensor("lokr_w2_b")
        t2 = state_dict.get_tensor("lokr_t2", raise_on_missing=False)

        if t2 is not None:
            w2 = rebuild_tucker(t2, w2a, w2b)
        else:
            # flatten for matmul if conv
            w2_b_flat = w2b.flatten(1) if w2b.dim() > 1 else w2b
            w2 = w2a @ w2_b_flat

        lora_dim = w2b.shape[0]

    while w1.dim() < w2.dim():
        w1 = w1.unsqueeze(-1)

    delta = torch.kron(w1, w2)

    alpha = state_dict.get_tensor("alpha", raise_on_missing=False)
    if alpha is not None and alpha.isfinite() and lora_dim is not None:
        delta *= alpha / lora_dim

    delta = delta.reshape(target_shape)
    return delta


def rebuild_tucker(t, wa, wb):
    rebuild2 = torch.einsum("i j ..., i p, j r -> p r ...", t, wa, wb)
    return rebuild2


class LycorisModelConfig(LazyModelConfigBase):
    def __init__(self, base_config: ModelConfig, lycoris_identifier: str, prefix: str, algorithms: Iterable[str]):
        super().__init__()
        self.base_config = base_config
        self.lycoris_identifier = lycoris_identifier
        self.prefix = prefix
        self.algorithms = list(sorted(algorithms))
        self.lycoris_to_base_keys = {
            lycoris_key: key
            for key, meta in base_config.keys().items()
            for lycoris_key in _to_lycoris_keys({key: dataclasses.replace(meta, shape=None)}, algorithms, self.prefix)
        }

    @property
    def identifier(self) -> str:
        return f"{self.base_config.identifier}_{self.lycoris_identifier}_lora"

    def create_config(self):
        identifier = self.identifier
        components = {
            k: _to_lycoris_keys(component.keys(), self.algorithms, self.prefix)
            for k, component in self.base_config.components().items()
        }
        return ModelConfigImpl(identifier, components)

    def to_lycoris_keys(self, key: StateDictKey, algos: Iterable[str] = None) -> Mapping[StateDictKey, KeyMetadata]:
        return _to_lycoris_keys({key: dataclasses.replace(self.base_config.keys()[key], shape=None)}, algos if algos is not None else self.algorithms, self.prefix)


def _to_lycoris_keys(
    base_keys: Mapping[StateDictKey, KeyMetadata],
    algorithms: Iterable[str],
    prefix: str,
) -> Dict[StateDictKey, KeyMetadata]:
    lycoris_keys = {}

    for key, meta in base_keys.items():
        for algorithm in algorithms:
            if key.endswith("bias") or not getattr(meta.dtype, "is_floating_point", True) or meta.optional:
                continue

            key = key.split('.')
            if key[-1] == "weight":
                key = key[:-1]
            key = "_".join(key)

            for suffix in lycoris_algorithms[algorithm]:
                lycoris_key = f"{prefix}_{key}.{suffix}"
                lycoris_keys[lycoris_key] = dataclasses.replace(meta, shape=[] if suffix == "alpha" else None, optional=True)

    return lycoris_keys


lycoris_algorithms = {
    "lora": ("lora_up.weight", "lora_mid.weight", "lora_down.weight", "alpha"),
    "lokr": ("lokr_w1", "lokr_w1_a", "lokr_w1_b", "lokr_w2", "lokr_w2_a", "lokr_w2_b", "lokr_t2", "alpha"),
}


_register_all_lycoris_configs()



import contextlib
import dataclasses
import threading
from collections import defaultdict
from sd_mecha.recipe_nodes import LiteralRecipeNode, MergeRecipeNode, ModelRecipeNode, RecipeNode, RecipeVisitor
from typing import Any, Dict, Mapping, Optional, Set, Tuple


def create_merge_method_context(recipe: RecipeNode, active_keys: Mapping[RecipeNode, Set[str]], enable_outputs_reuse: bool) -> Dict[RecipeNode, "MergeMethodContext"]:
    ports_visitor = GetOutputPortsVisitor(active_keys)
    ports_visitor.visit_root(recipe)
    parent_ports = ports_visitor.parent_ports
    create_context_visitor = CreateMergeMethodContextVisitor(active_keys, parent_ports, enable_outputs_reuse)
    res = recipe.accept(create_context_visitor)
    return res


@dataclasses.dataclass
class GetOutputPortsVisitor(RecipeVisitor):
    active_keys: Mapping[RecipeNode, Set[str]]
    parent_ports: Dict[RecipeNode, Dict[str, Set[Optional[Tuple[RecipeNode, str, int | str]]]]] = dataclasses.field(default_factory=lambda: defaultdict(lambda: defaultdict(set)))

    def visit_root(self, root: RecipeNode):
        for k in self.active_keys[root]:
            self.parent_ports[root][k].add(None)
        root.accept(self)

    def visit_literal(self, node: LiteralRecipeNode):
        for child_node in set(node.value_dict.values()):
            if isinstance(child_node, RecipeNode):
                child_node.accept(self)

    def visit_model(self, node: ModelRecipeNode):
        pass

    def visit_merge(self, node: MergeRecipeNode):
        param_names = node.merge_method.get_param_names().as_dict(len(node.bound_args.args))
        key_map = node.key_map()

        for output_key in self.active_keys[node]:
            if output_key not in key_map:
                continue

            for input_idx, param_name in param_names.items():
                child_node = node.bound_args.args[input_idx] if isinstance(input_idx, int) else node.bound_args.kwargs[input_idx]

                child_output_keys = self.active_keys[child_node].intersection(key_map[output_key].inputs[param_name])
                for child_output_key in child_output_keys:
                    self.parent_ports[child_node][child_output_key].add((node, output_key, input_idx))

        for input_node in (*node.bound_args.args, *node.bound_args.kwargs.values()):
            input_node.accept(self)


@dataclasses.dataclass
class CreateMergeMethodContextVisitor(RecipeVisitor):
    active_keys: Mapping[RecipeNode, Set[str]]
    parent_ports: Mapping[RecipeNode, Mapping[str, Set[Optional[Tuple[RecipeNode, str, int | str]]]]]
    enable_outputs_reuse: bool

    def visit_literal(self, node: LiteralRecipeNode) -> Dict[RecipeNode, "MergeMethodContext"]:
        res = {}
        for key, nested_node in node.value_dict.items():
            if isinstance(nested_node, RecipeNode):
                res |= nested_node.accept(self)
        return res

    def visit_model(self, node: ModelRecipeNode) -> Dict[RecipeNode, "MergeMethodContext"]:
        return {}

    def visit_merge(self, node: MergeRecipeNode) -> Dict[RecipeNode, "MergeMethodContext"]:
        res = {}
        for input_node in (*node.bound_args.args, *node.bound_args.kwargs.values()):
            res |= input_node.accept(self)

        if not node.merge_method.reuse_outputs:
            res[node] = MergeMethodContext({}, node.merge_method.instantiate(), reused_output_keys=self.active_keys[node])
            return res

        key_map = node.key_map()

        node_parent_ports = self.parent_ports[node]
        locks_map = defaultdict(lambda: defaultdict(threading.Lock if self.enable_outputs_reuse else contextlib.nullcontext))

        refs = {
            output_key: MergeMethodOutputRef(parent_ports, None, locks_map[node][key_map[output_key].outputs])
            for output_key, parent_ports in node_parent_ports.items()
            if (
                len(parent_ports) > 1 or
                len(parent_ports) > 0 and len(key_map[output_key].outputs) > 1
            )
        }
        res[node] = MergeMethodContext(
            refs if self.enable_outputs_reuse else {},
            node.merge_method.instantiate(),
            self.active_keys[node] if self.enable_outputs_reuse else (self.active_keys[node] - set(refs)),
        )

        return res


@dataclasses.dataclass
class MergeMethodContext:
    output_refs: Dict[str, "MergeMethodOutputRef"]
    instance: Any
    reused_output_keys: Set[str]

    @contextlib.contextmanager
    def output_ref_context(self, key: str, lock: bool = True) -> "MergeMethodOutputRef":
        output_ref = self.output_refs.get(key)
        if output_ref is None:
            yield MergeMethodOutputRef(set(), None, contextlib.nullcontext(), locked=True)
        elif lock:
            with output_ref:
                yield output_ref
            if output_ref.was_freed():
                self.output_refs.pop(key, None)
        else:
            with output_ref.assumed_locked():
                yield output_ref


@dataclasses.dataclass
class MergeMethodOutputRef:
    remaining_ports: Set[Optional[Tuple[RecipeNode, str]]]
    cache: Any
    lock: contextlib.AbstractContextManager
    locked: bool = False

    def use_once(self, port: Optional[Tuple[RecipeNode, str]]) -> Any:
        assert self.locked
        res = self.cache
        self.remaining_ports.discard(port)
        if self.was_freed():
            self.cache = None
        return res

    def was_freed(self) -> bool:
        return len(self.remaining_ports) <= 0

    def set_cache(self, cache: Any):
        assert self.locked
        if self.was_freed():
            return
        assert self.cache is None, f"cache was already set: {self.cache}, trying to assign {cache}"
        self.cache = cache

    @contextlib.contextmanager
    def assumed_locked(self):
        old = self.locked
        self.locked = True
        try:
            yield self
        finally:
            self.locked = old

    def __enter__(self):
        self.lock.__enter__()
        self.locked = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.locked = False
        self.lock.__exit__(exc_type, exc_val, exc_tb)



import torch
import sd_mecha
from sd_mecha import merge_method, Parameter, Return, StateDict
from sd_mecha.extensions import model_configs
from sd_mecha.merge_context import MergeMethodContext
from sd_mecha.recipe_nodes import RecipeNode
from torch import Tensor
from typing import Dict
from unittest.mock import patch


dtype = torch.float32
test_config = model_configs.ModelConfigImpl(
    "test1-mecha", {
        "component": model_configs.ModelComponent({
            "key": model_configs.KeyMetadata([2, 2], dtype),
        })
    }
)
model_configs.register(test_config)


test_config2 = model_configs.ModelConfigImpl(
    "test2-mecha", {
        "component": model_configs.ModelComponent({
            "key.0": model_configs.KeyMetadata([2], dtype),
            "key.1": model_configs.KeyMetadata([2], dtype),
        })
    }
)
model_configs.register(test_config2)


@merge_method(is_conversion=True)
class convert_test_to_test2:
    @staticmethod
    def map_keys(b):
        b["key.0", "key.1"] = b.keys["key"]

    def __call__(
        self,
        a: Parameter(StateDict[Tensor], model_config=test_config),
        **kwargs,
    ) -> Return(StateDict[Tensor], model_config=test_config2):
        value = a["key"]
        return {
            "key.0": value[0],
            "key.1": value[1],
        }


@merge_method(is_conversion=True)
class convert_test2_to_test:
    @staticmethod
    def map_keys(b):
        b["key"] = b.keys["key.0", "key.1"]

    def __call__(
        self,
        a: Parameter(StateDict[Tensor], model_config=test_config2),
        **kwargs,
    ) -> Return(Tensor, model_config=test_config):
        value0 = a["key.0"]
        value1 = a["key.1"]
        return torch.stack([value0, value1], dim=0)


@merge_method(register=False)
def weight_delta(
    a: Parameter(Tensor, merge_space="delta"),
    alpha: Parameter(float) = 1.0,
) -> Return(Tensor, merge_space="delta"):
    return a * alpha


def make_value(value, merge_space=None):
    return sd_mecha.literal({"key": torch.tensor([[value, value], [value, value]], dtype=dtype)}, config=test_config, merge_space=merge_space)


def make_value2(value, merge_space=None):
    return sd_mecha.literal({"key.0": torch.tensor([value, value], dtype=dtype), "key.1": torch.tensor([value, value], dtype=dtype)}, config=test_config2, merge_space=merge_space)


class InterceptReturnPatch:
    def __init__(self, original_fn):
        self.original_fn = original_fn
        self.return_value = None

    def __call__(self, *args, **kwargs) -> Dict[RecipeNode, MergeMethodContext]:
        self.return_value = self.original_fn(*args, **kwargs)
        return self.return_value


def patch_create_context():
    create_merge_method_context_patch = InterceptReturnPatch(sd_mecha.merge_context.create_merge_method_context)
    return patch("sd_mecha.merging.create_merge_method_context", new=create_merge_method_context_patch)


def call_merge(recipe):
    sd_mecha.merge(
        recipe,
        strict_weight_space=False,
        threads=0,
        merge_device=None,
        merge_dtype=None,
        output_device=None,
        output_dtype=None,
    )


def act_assert_no_leak(recipe):
    with patch_create_context() as create_context_patch:
        call_merge(recipe)

    for node, mm_context in create_context_patch.return_value.items():
        num_leaked = sum(not output_ref.was_freed() for output_ref in mm_context.output_refs.values())
        assert num_leaked == 0


def test_simple_tree():
    a = make_value(5)
    b = make_value(2)
    recipe = sd_mecha.weighted_sum(a, b)

    act_assert_no_leak(recipe)


def test_binary_tree():
    a = make_value(5)
    b = make_value(2)
    c = make_value(3)
    d = make_value(9)
    m1 = sd_mecha.weighted_sum(a, b)
    m2 = sd_mecha.weighted_sum(c, d)
    recipe = sd_mecha.weighted_sum(m1, m2)

    act_assert_no_leak(recipe)


def test_simple_dag():
    a = make_value(5, "delta")
    b = sd_mecha.truncate_rank(a, rank_ratio=0.5)
    c = sd_mecha.truncate_rank(b, rank_ratio=0.25)
    d = sd_mecha.truncate_rank(b, rank_ratio=0.3)
    recipe = sd_mecha.weighted_sum(c, d)

    act_assert_no_leak(recipe)


def test_back_and_forth_conversion():
    a = make_value(5)
    b = sd_mecha.convert(a, test_config2)
    recipe = sd_mecha.convert(b, test_config)

    act_assert_no_leak(recipe)


def test_back_and_transform_and_forth_conversion():
    a = make_value(5, "delta")
    b = sd_mecha.convert(a, test_config2)
    c = sd_mecha.weighted_sum(b, weight_delta(b, 0.5))
    d = sd_mecha.convert(c, test_config)

    act_assert_no_leak(d)



import torch
from torch import Tensor
from sd_mecha.extensions.merge_methods import merge_method, Parameter, Return


@merge_method
def add_cosine_a(
    a: Parameter(Tensor, "weight"),
    b: Parameter(Tensor, "weight"),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "weight"):
    a_norm = torch.nn.functional.normalize(a, dim=0)
    b_norm = torch.nn.functional.normalize(b, dim=0)
    similarity = torch.nn.functional.cosine_similarity(a_norm, b_norm, dim=0)
    return add_cosine_generic(a, b, alpha, similarity)


@merge_method
def add_cosine_b(
    a: Parameter(Tensor, "weight"),
    b: Parameter(Tensor, "weight"),
    alpha: Parameter(Tensor) = 1.0,
) -> Return(Tensor, "weight"):
    similarity = torch.nn.functional.cosine_similarity(a, b, dim=0)
    dot_product = torch.sum(a * b)
    magnitude_similarity = dot_product / (torch.norm(a) * torch.norm(b))
    combined_similarity = (similarity + magnitude_similarity) / 2.0
    return add_cosine_generic(a, b, alpha, combined_similarity)


def add_cosine_generic(a: Tensor, b: Tensor, alpha: Tensor, similarity: Tensor) -> Tensor:
    k = 1 - torch.clamp(similarity - alpha, 0, 1)
    return torch.lerp(a, b, k)



import contextlib
import dataclasses
import functools
import logging
import operator
import os
import pathlib
import sys
import threading
import typing
import torch
import sd_mecha
from .extensions import model_dirs
from sd_mecha.merge_context import create_merge_method_context, MergeMethodContext
from .extensions.merge_methods import value_to_node, StateDict, T as MergeMethodT
from .extensions.merge_spaces import MergeSpace
from .extensions.model_configs import ModelConfig, KeyMetadata
from .graph_finalization import open_graph
from .keys_map import KeyRelation
from .recipe_nodes import RecipeVisitor, LiteralRecipeNode, RecipeNode, MergeRecipeNode, ModelRecipeNode, RecipeNodeOrValue, NonDictLiteralValue
from .streaming import OutSafetensorsDict, TensorMetadata, StateDictKeyError
from . import recipe_nodes, serialization
from collections import defaultdict, OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from contextlib import nullcontext
from tqdm import tqdm as tqdm_original
from types import SimpleNamespace
from typing import Optional, Mapping, MutableMapping, List, Iterable, Tuple, Dict, TypeVar, Sequence, Any
from .typing_ import is_subclass


def merge(
    recipe: RecipeNodeOrValue,
    *,
    fallback_model: Optional[RecipeNodeOrValue] = ...,
    merge_device: Optional[str | torch.device] = ...,
    merge_dtype: Optional[torch.dtype] = ...,
    output_device: Optional[str | torch.device] = ...,
    output_dtype: Optional[torch.dtype] = ...,
    threads: Optional[int] = ...,
    total_buffer_size: int = ...,
    strict_weight_space: bool = ...,
    check_finite: bool = ...,
    check_finite_mandatory_inputs: bool = ...,
    check_finite_optional_inputs: bool = ...,
    omit_extra_keys: bool = ...,
    omit_ema: bool = ...,
    check_mandatory_keys: bool = ...,
    tqdm: type = ...,
    validate_mm_contract: bool = ...,
    cache: Dict[RecipeNode, dict] = ...,
    reuse_outputs: bool = ...,
    output: Optional[MutableMapping[str, torch.Tensor]] | pathlib.Path | str = ...,
) -> Optional[MutableMapping[str, torch.Tensor]]:
    """
    Materialize a state dict from a recipe graph and optionally save it to a file.

    For each key of the target model config, execute all instructions of the recipe graph
    and store the result into a dictionary using the specified output strategy.

    Args:
        recipe:
            A `RecipeNode`, python literal, or dictionary describing how to merge or transform multiple models.
        fallback_model (optional):
            A secondary recipe or model to provide values for any keys missing from `recipe`.
        merge_device (optional):
            Device to load intermediate tensors onto while merging (e.g., "cpu" or "cuda").
        merge_dtype (optional):
            Torch dtype for intermediate merges (e.g., `torch.float32`, `torch.float64`).
        output_device (optional):
            Final output device (e.g., "cpu").
        output_dtype (optional):
            Final dtype for the saved model.
        threads (optional):
            Number of threads to spawn for parallel merges. Defaults to a reasonable guess.
        total_buffer_size (optional):
            Total byte size of the buffers for all safetensors state dicts (input and output).
        strict_weight_space (optional):
            If True, verifies that merges occur in "weight" space. If False, merges can happen
            in other merge spaces (like "delta" or "param"). Defaults to True.
        check_finite (optional):
            If True, warns if any non-finite values appear in the output model. Defaults to True.
        check_finite_mandatory_inputs (optional):
            If True, automatically discards (raises StateDictKeyError) directly from model nodes for mandatory key containing non-finite values. Defaults to False.
        check_finite_optional_inputs (optional):
            If True, automatically discards (raises StateDictKeyError) directly from model nodes for optional key containing non-finite values. Defaults to True.
        omit_extra_keys (optional):
            If True, warns about and removes unrecognized keys from the output model. Defaults to True.
        omit_ema (optional):
            If True, omits ema keys from the output model. Defaults to omit_extra_keys.
        check_mandatory_keys (optional):
            If True and an input model is missing non-optional keys, raises RuntimeError. Defaults to False.
        tqdm (optional):
            A custom progress-bar factory. By default, uses `tqdm.tqdm`.
        output (optional):
            Where to store the merged state dict. Can be a filesystem path (string or
            `Path`) ending with `.safetensors`, an in-memory dict-like object, or None.
            If it is None or omitted, an empty dict is created and returned when the merge completes.
        validate_mm_contract (optional):
            If True, validates that merge methods return the right amount of outputs indicated by `map_keys`
            and do not read other inputs than those reported by `map_keys`
        cache (optional):
            Dictionary of caches for each recipe node in `recipe`. Each dict should be empty on the first call.
            If set, the dicts are filled by the respective merge methods on the first call to merge(), and then reused for subsequent calls to merge().
            This can speed up certain merge methods when testing multiple parameter variations with fixed inputs.
        reuse_outputs (optional):
            If True, temporarily memoizes the output of merge nodes that are used by multiple parents to avoid recomputations.
            To minimize extra memory usage, each memoized output is freed as soon as all its consumers consumed it. Defaults to True.

    Returns:
        The in-memory dictionary if `output` is either a MutableMapping or None, and nothing if `output` is a file path.
    """
    if output is ...:
        output = None
    if fallback_model is ...:
        fallback_model = None
    if merge_device is ...:
        merge_device = "cpu"
    if merge_dtype is ...:
        merge_dtype = torch.float32
    cast_inputs = merge_device is not None and merge_dtype is not None
    if output_device is ...:
        output_device = "cpu"
    if output_dtype is ...:
        output_dtype = torch.float16
    if threads is ...:
        threads = None
    if total_buffer_size is ...:
        total_buffer_size = 2**28
    if strict_weight_space is ...:
        strict_weight_space = True
    if check_finite is ...:
        check_finite = True
    if check_finite_mandatory_inputs is ...:
        check_finite_mandatory_inputs = False
    if check_finite_optional_inputs is ...:
        check_finite_optional_inputs = True
    check_finite_inputs = check_finite_mandatory_inputs or check_finite_optional_inputs
    if omit_extra_keys is ...:
        omit_extra_keys = True
    if omit_ema is ...:
        omit_ema = omit_extra_keys
    if check_mandatory_keys is ...:
        check_mandatory_keys = False
    if tqdm is ...:
        tqdm = tqdm_original
    if validate_mm_contract is ...:
        validate_mm_contract = True
    if cache is ...:
        cache = {}
    if reuse_outputs is ...:
        reuse_outputs = True

    if threads is not None and (not isinstance(threads, int) or threads < 0):
        raise RuntimeError("threads should be a non-negative integer or None")

    recipe = value_to_node(recipe)
    original_recipe = recipe
    if cast_inputs or check_finite_inputs:
        cast_visitor = CastInputDicts(merge_device, merge_dtype, check_finite_mandatory_inputs, check_finite_optional_inputs)
        recipe = recipe.accept(cast_visitor)
        cache = {cast_visitor.converted_nodes.get(node, node): cache_dict for node, cache_dict in cache.items()}

    if fallback_model is not None:
        recipe = recipe | fallback_model

    if output_device is not None or output_dtype is not None:
        recipe = recipe.to(device=output_device, dtype=output_dtype)

    total_files_open = (
        recipe.accept(recipe_nodes.ModelsCountVisitor()) +
        int(isinstance(output, (str, pathlib.Path)))
    )
    buffer_size_per_file = total_buffer_size // max(1, total_files_open)
    if threads is None:
        threads = min(max(total_files_open, 2), os.cpu_count() or 0, 4)

    if threads == 0:
        thread_local_data = SimpleNamespace()
        executor = ThisThreadExecutor()
    else:
        thread_local_data = threading.local()
        executor = ThreadPoolExecutor(max_workers=threads)

    with open_graph(
        recipe,
        buffer_size_per_file,
    ) as graph:
        finalized_res = graph.finalize(
            check_extra_keys=omit_extra_keys,
            check_mandatory_keys=check_mandatory_keys,
            model_config_preference=("singleton-mecha",),
            merge_space="weight" if strict_weight_space else None,
            merge_space_preference=("weight",) if not strict_weight_space else None,
        )
        recipe, node_to_keys = finalized_res.root, finalized_res.node_to_keys
        minimal_keys = node_to_keys[recipe]
        cache = {finalized_res.to_finalized_node[node]: cache_dict for node, cache_dict in cache.items()}

        model_config = recipe.model_config
        graph_metadata = {k: v for k, v in recipe.model_config.keys().items() if k in minimal_keys}
        if not omit_extra_keys:
            recipe, graph_metadata = copy_extra_keys(recipe)

        if omit_ema and "ema" in model_config.components():
            for key in model_config.components()["ema"].keys():
                if key in graph_metadata:
                    # remove ema keys from merge plan
                    del graph_metadata[key]

        buffer_size_per_file_per_thread = buffer_size_per_file // max(1, threads)
        merge_methods_context = create_merge_method_context(recipe, node_to_keys, enable_outputs_reuse=reuse_outputs)

        with (
            executor,
            tqdm(total=len(graph_metadata), desc="Merging recipe") as progress,
            _get_output_dict(
                output,
                graph_metadata,
                original_recipe,
                buffer_size_per_file_per_thread,
            ) as output_dict,
        ):
            fix_torch_threading()
            futures = []
            for key, key_metadata in graph_metadata.items():
                fn = recipe.accept
                fn = _track_output(fn, output_dict, key, key_metadata, check_finite)
                fn = _track_progress(fn, key, graph_metadata[key].shape, progress)
                fn = _wrap_thread_context(fn, thread_local_data)
                futures.append(executor.submit(fn, KeyMergeVisitor(key, merge_methods_context, validate_mm_contract, cache)))

            for future in as_completed(futures):
                if future.exception() is not None:
                    for future_to_cancel in futures:
                        future_to_cancel.cancel()
                future.result()

            for node, mm_context in merge_methods_context.items():
                num_leaked = sum(not output_ref.was_freed() for output_ref in mm_context.output_refs.values())
                if num_leaked:
                    logging.warning(f"memory leaked during the merge: {node}, number of entries: {num_leaked}")

            if output is None:
                return output_dict


def fix_torch_threading():
    if torch.cuda.is_available():
        # this greedy loads the torch.linalg module
        # avoids a hard error caused by threads>1 with some torch ops
        # see https://github.com/pytorch/pytorch/issues/90613
        torch.linalg.inv(torch.ones((1, 1), device="cuda"))

    globals()['fix_torch_threading'] = lambda: None


@dataclasses.dataclass
class MinimalMetadataVisitor(RecipeVisitor):
    def visit_literal(self, node: LiteralRecipeNode) -> Mapping[str, KeyMetadata]:
        meta = node.model_config.keys()
        res = {}

        child_keys = defaultdict(set)
        for k, child_node in node.value_dict.items():
            if isinstance(child_node, RecipeNode):
                child_keys[child_node].add(k)

        for k, child_node in node.value_dict.items():
            if isinstance(child_node, RecipeNode):
                child_res = child_node.accept(dataclasses.replace(self))
                for k_res in child_keys[child_node]:
                    child_keys[k_res] = child_res[k_res]
            else:
                res[k] = meta[k]

        return res

    def visit_model(self, node: ModelRecipeNode) -> Mapping[str, KeyMetadata]:
        return {k: v for k, v in node.state_dict.metadata()}

    def visit_merge(self, node: MergeRecipeNode) -> Mapping[str, KeyMetadata]:
        res = {}
        meta = node.model_config.keys()
        key_map = node.key_map()

        for param, args in node.all_args().items():
            for arg in (args if isinstance(args, tuple) else (args,)):
                arg_res = arg.accept(self)
                for input_key in arg_res:
                    output_keys = key_map.in_to_out.get(input_key)
                    output_keys = output_keys.get(param) if output_keys is not None else None
                    if output_keys is None:
                        continue

                    for output_key in output_keys:
                        res[output_key] = meta[output_key]

        return res


def copy_extra_keys(recipe: RecipeNode):
    config = recipe.model_config
    merge_space = recipe.merge_space
    metadata = config.keys()
    aliases = {k_alias: k for k, k_aliases in config.aliases().items() for k_alias in k_aliases}

    forwardable_nodes_visitor = ForwardableNodesVisitor(config, merge_space)
    recipe.accept(forwardable_nodes_visitor)
    forwardable_nodes = forwardable_nodes_visitor.forwardable_nodes

    new_recipe = functools.reduce(operator.or_, [n[0] for n in forwardable_nodes], recipe)
    new_metadata = OrderedDict((aliases.get(k, k), v) for n in forwardable_nodes for k, v in n[1].items()) | metadata
    return new_recipe, new_metadata


@dataclasses.dataclass
class ForwardableNodesVisitor(RecipeVisitor):
    target_config: ModelConfig
    target_merge_space: MergeSpace
    forwardable_nodes: List[Tuple[RecipeNode, Mapping[str, KeyMetadata]]] = dataclasses.field(default_factory=list)

    def visit_literal(self, node: LiteralRecipeNode):
        if not node.value_dict:
            return

        can_forward = (
            node.model_config == self.target_config and
            node.merge_space == self.target_merge_space and
            not any(node in n[0] for n in self.forwardable_nodes)
        )
        metadata = {}
        for k, v in node.value_dict.items():
            if isinstance(v, RecipeNode):
                v.accept(self)
            elif can_forward and isinstance(v, torch.Tensor):
                metadata[k] = KeyMetadata(v.shape, v.dtype)

        if can_forward and metadata:
            self.forwardable_nodes.append((node, metadata))

    def visit_model(self, node: ModelRecipeNode):
        if node.model_config == self.target_config and not any(node in n[0] for n in self.forwardable_nodes):
            self.forwardable_nodes.append((
                node,
                OrderedDict(
                    (k, KeyMetadata(v.shape, v.dtype))
                    for k, v in node.state_dict.metadata()
                )
            ))

    def visit_merge(self, node: MergeRecipeNode):
        for arg in node.all_args().values():
            arg.accept(self)


@contextlib.contextmanager
def _get_output_dict(
    output: Optional[MutableMapping[str, torch.Tensor]] | pathlib.Path | str,
    merged_header: Mapping[str, KeyMetadata],
    recipe: RecipeNode,
    buffer_size_per_thread: int,
):
    if isinstance(output, (str, pathlib.Path)):
        if not isinstance(output, pathlib.Path):
            output = pathlib.Path(output)
        if not output.is_absolute():
            for model_dir in model_dirs.get_all():
                output = model_dir / output
                break
        logging.info(f"Saving to {output}")

        try:
            serialized_recipe = serialization.serialize(recipe)
        except TypeError:
            logging.warning("The recipe graph could not be serialized. The output state dict will not contain the recipe.")
            serialized_recipe = None
        streamed_output = OutSafetensorsDict(
            output,
            OrderedDict((k, TensorMetadata(v.shape, v.dtype)) for k, v in merged_header.items()),
            serialized_recipe,
            buffer_size_per_thread,
        )
        try:
            yield streamed_output
        finally:
            streamed_output.close()
    else:
        if output is None:
            output = {}
        yield output


def _track_output(fn, output, key: str, key_metadata: KeyMetadata, check_finite: bool):
    @functools.wraps(fn)
    def track_output(*args, **kwargs):
        try:
            res = fn(*args, **kwargs)
            if check_finite and isinstance(res, torch.Tensor):
                try:
                    all_finite = res.isfinite().all()
                except RuntimeError:  # for example, fp8_e4m3fn doesn't support .isfinite()
                    all_finite = res.to(dtype=torch.bfloat16).isfinite().all()

                if not all_finite:
                    logging.warning(f"there are non finite values in key '{key}': {key_metadata}")

            output[key] = res
        except StateDictKeyError as e:
            if key_metadata.optional:
                logging.debug(f"skipping key {e}")
            else:
                raise RuntimeError(f"could not merge mandatory key: {e}") from e
    return track_output


def _track_progress(fn, key, key_shape, progress):
    @functools.wraps(fn)
    def track_progress(*args, **kwargs):
        progress.set_postfix({"key": key} | ({"shape": list(key_shape)} if key_shape is not None else {}))
        res = fn(*args, **kwargs)
        progress.update()
        return res
    return track_progress


def _wrap_thread_context(fn, ctx):
    @functools.wraps(fn)
    def thread_context(*args, **kwargs):
        if torch.cuda.is_available():
            if not hasattr(ctx, 'cuda_stream'):
                setattr(ctx, "cuda_stream", torch.cuda.Stream())
            with torch.cuda.stream(ctx.cuda_stream):
                return fn(*args, **kwargs)
        else:
            return fn(*args, **kwargs)

    return thread_context


class ThisThreadExecutor(nullcontext):
    def submit(self, fn, /, *args, **kwargs):
        result = Future()
        try:
            result.set_result(fn(*args, **kwargs))
        except BaseException as e:
            result.set_exception(e)
        return result


@dataclasses.dataclass
class KeyMergeVisitor(RecipeVisitor):
    output_key: str
    merge_methods_context: Dict[RecipeNode, MergeMethodContext]
    validate_mm_contract: bool
    merge_methods_caches: Dict[RecipeNode, Any]
    parent_port: Optional[Tuple[RecipeNode, str, int | str]] = None

    def visit_literal(self, node: LiteralRecipeNode):
        value = self.__visit_sd(node.value_dict, node.model_config)
        if isinstance(value, NonDictLiteralValue):
            return value
        if isinstance(value, RecipeNode):
            return value.accept(self)
        raise TypeError(f"Unexpected literal node value of type {type(value)}")

    def visit_model(self, node: recipe_nodes.ModelRecipeNode) -> torch.Tensor:
        return self.__visit_sd(node.state_dict, node.model_config)

    def __visit_sd(self, state_dict, model_config):
        aliases = model_config.aliases().get(self.output_key, [])
        possible_keys = [self.output_key, *aliases]

        for i, possible_key in enumerate(possible_keys):
            try:
                return state_dict[possible_key]
            except KeyError as e:
                if i == len(possible_keys) - 1:
                    raise StateDictKeyError(self.output_key) from e

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode) -> torch.Tensor:
        context = self.merge_methods_context[node]
        with context.output_ref_context(self.output_key) as output_ref:
            if output_ref.cache is not None:
                res = output_ref.use_once(self.parent_port)
            else:
                try:
                    key_map = node.key_map()
                    assert self.output_key in key_map, f"Merge method {node.merge_method} does not produce key {self.output_key}."
                    key_relation = key_map[self.output_key]

                    merged_args, merged_kwargs = self.__visit_deeper_first(node, key_relation)
                    res = node.merge_method.merge_key(
                        merged_args,
                        merged_kwargs,
                        self.output_key,
                        key_relation,
                        self.merge_methods_caches.get(node),
                        context.instance,
                        self.output_key in context.reused_output_keys,
                    )
                finally:
                    if self.parent_port is None:
                        release_visitor = KeyReleaseVisitor(self.output_key, self.merge_methods_context, self.parent_port, needs_lock=False)
                        node.accept(release_visitor)
                if isinstance(res, dict):
                    if self.validate_mm_contract:
                        assert set(key_relation.outputs) == set(res.keys()), (
                            f"Merge method {node.merge_method.identifier} returned an unexpected set of keys: {list(res)}",
                        )
                    for key, value in res.items():
                        with context.output_ref_context(key, lock=False) as sibling_output_ref:
                            # output_ref.cache is set here
                            sibling_output_ref.set_cache(value)
                    res = res[self.output_key]
                else:
                    output_ref.set_cache(res)

        return res

    def __visit_deeper_first(
        self,
        node: MergeRecipeNode,
        key_relation: KeyRelation,
    ) -> Tuple[
        Sequence[NonDictLiteralValue | StateDict[NonDictLiteralValue]],
        Mapping[str, NonDictLiteralValue | StateDict[NonDictLiteralValue]],
    ]:
        def depth_of_value(index) -> int:
            nodes = node.bound_args.args if isinstance(index, int) else node.bound_args.kwargs
            return nodes[index].accept(recipe_nodes.ModelDepthRecipeVisitor())

        error_holder = ErrorHolder()
        merged = {}
        input_types = node.merge_method.get_input_types().as_dict(len(node.bound_args.args))
        input_names = node.merge_method.get_param_names().as_dict(len(node.bound_args.args))
        indices = (*range(len(node.bound_args.args)), *node.bound_args.kwargs.keys())

        for index in sorted(indices, key=depth_of_value, reverse=True):
            parent_port = (node, self.output_key, index)
            input_name = input_names[index]
            input_node = node.bound_args.args[index] if isinstance(index, int) else node.bound_args.kwargs[index]
            input_visitor = dataclasses.replace(self, parent_port=parent_port)
            if is_subclass(input_types[index], StateDict):
                if self.validate_mm_contract:
                    input_keys_constraints = key_relation.inputs[input_name]
                else:
                    input_keys_constraints = None
                expected_type = next(iter(typing.get_args(input_types[index]) or (MergeMethodT,)))
                merged[index] = error_holder.intercept(MergeNodeWrapperStateDict, input_node, expected_type, input_visitor, input_keys_constraints)
            else:
                merged[index] = cast_node_value(error_holder.intercept(input_node.accept, input_visitor), input_types[index])

        merged_args = tuple(merged.get(index) for index in range(len(node.bound_args.args)))
        merged_kwargs = {k: v for k, v in merged.items() if not isinstance(k, int)}
        error_holder.try_raise()
        return merged_args, merged_kwargs


@dataclasses.dataclass
class KeyReleaseVisitor(RecipeVisitor):
    output_key: str
    merge_methods_context: Dict[RecipeNode, MergeMethodContext]
    parent_port: Optional[Tuple[RecipeNode, str, int | str]]
    needs_lock: bool

    def visit_literal(self, node: LiteralRecipeNode):
        value = self.__visit_sd(node.value_dict, node.model_config)
        if isinstance(value, RecipeNode):
            value.accept(self)

    def visit_model(self, node: recipe_nodes.ModelRecipeNode):
        pass

    def __visit_sd(self, state_dict, model_config):
        aliases = model_config.aliases().get(self.output_key, [])
        keys = [self.output_key, *aliases]

        for i, key in enumerate(keys):
            try:
                return state_dict[key]
            except KeyError:
                pass

        return None

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode):
        context = self.merge_methods_context[node]
        with context.output_ref_context(self.output_key, lock=self.needs_lock) as output_ref:
            output_ref.use_once(self.parent_port)
            self.__visit_deeper_first(node)

    def __visit_deeper_first(
        self,
        node: MergeRecipeNode,
    ):
        error_holder = ErrorHolder()

        key_map = node.key_map()
        if self.output_key not in key_map:
            return

        key_relation = key_map[self.output_key]

        for input_idx, input_param in node.merge_method.get_param_names().as_dict(len(node.bound_args.args)).items():
            input_node = node.bound_args.args[input_idx] if isinstance(input_idx, int) else node.bound_args.kwargs[input_idx]
            parent_port = (node, self.output_key, input_idx)
            release_visitors = [
                KeyReleaseVisitor(input_key, self.merge_methods_context, parent_port, needs_lock=True)
                for input_key in key_relation.inputs[input_param]
            ]
            for release_visitor in release_visitors:
                error_holder.intercept(input_node.accept, release_visitor)

        error_holder.try_raise()


class ErrorHolder:
    def __init__(self):
        self.exc_info = None

    def intercept(self, fn, *args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:
            self.exc_info = sys.exc_info()
            return None

    def try_raise(self):
        if self.exc_info:
            raise self.get_error()

    def get_error(self):
        if self.exc_info:
            return self.exc_info[1].with_traceback(self.exc_info[2])
        else:
            return None


class MergeNodeWrapperStateDict(StateDict):
    def __init__(
        self,
        node: recipe_nodes.RecipeNode,
        expected_type: type,
        original_merge_visitor: KeyMergeVisitor,
        input_keys_constraints: Optional[Tuple[str, ...]],
    ):
        self.node = node
        self.expected_type = expected_type
        self.original_merge_visitor = original_merge_visitor
        self.input_keys_constraints = input_keys_constraints

    def __getitem__(self, key):
        if self.input_keys_constraints is not None:
            assert key in self.input_keys_constraints, (
                f"node {self.node} tried to fetch key '{key}' "
                f"but only requested these keys through input_keys_for_output(): {self.input_keys_constraints}"
            )

        key_merger = dataclasses.replace(self.original_merge_visitor, output_key=key)
        res_raw = self.node.accept(key_merger)
        res = cast_node_value(res_raw, self.expected_type)
        return res

    def __len__(self):
        return len(self.compute_keys())

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, item):
        return item in self.compute_keys()

    def keys(self) -> Iterable[str]:
        return self.compute_keys().keys()

    @property
    def model_config(self) -> ModelConfig:
        return self.node.model_config

    def compute_keys(self):
        return self.model_config.keys()


def cast_node_value(value, expected_type):
    if value is None:
        return value

    try:
        if issubclass(typing.get_origin(expected_type) or expected_type, StateDict):
            expected_type = (typing.get_args(expected_type) + (MergeMethodT,))[0]
    except TypeError:
        pass

    if isinstance(expected_type, TypeVar) or isinstance(value, expected_type):
        return value
    if isinstance(expected_type, str):
        raise RuntimeError(f"cannot implicitly convert {type(value)} to {expected_type}")
    if issubclass(expected_type, int):
        return int(value)
    if issubclass(expected_type, float):
        return float(value)
    if issubclass(expected_type, bool):
        return bool(value)
    if issubclass(expected_type, torch.Tensor):
        return torch.tensor(value, dtype=torch.float32)
    return value


@dataclasses.dataclass
class CastInputDicts(RecipeVisitor):
    device: str | torch.device
    dtype: torch.dtype
    check_finite_mandatory: bool
    check_finite_optional: bool
    converted_nodes: Dict[RecipeNode, RecipeNode] = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        self.check_finite = self.check_finite_mandatory or self.check_finite_optional

    def visit_literal(self, node: LiteralRecipeNode):
        if node in self.converted_nodes:
            return self.converted_nodes[node]

        converted_dict = {}
        can_check_finite = False
        for k, v in node.value_dict.items():
            if isinstance(v, RecipeNode):
                v = v.accept(self)
            elif isinstance(v, (int, float, bool, str)):
                v = v
            elif isinstance(v, torch.Tensor):
                v = v.to(device=self.device, dtype=self.dtype)
                can_check_finite = True
            else:
                raise RuntimeError(f"Cannot cast type {type(v)} to device={self.device}, dtype={self.dtype}")
            converted_dict[k] = v

        res = LiteralRecipeNode(converted_dict, node.model_config, node.merge_space)
        if can_check_finite and self.check_finite:
            res = sd_mecha.omit_non_finite(res, self.check_finite_mandatory, self.check_finite_optional)
        self.converted_nodes[node] = res
        return res

    def visit_model(self, node: ModelRecipeNode):
        if node in self.converted_nodes:
            return self.converted_nodes[node]

        res = node.to(device=self.device, dtype=self.dtype)
        if self.check_finite:
            res = sd_mecha.omit_non_finite(res, self.check_finite_mandatory, self.check_finite_optional)
        self.converted_nodes[node] = res
        return res

    def visit_merge(self, node: MergeRecipeNode):
        if node in self.converted_nodes:
            return self.converted_nodes[node]

        args = tuple(v.accept(self) for v in node.bound_args.args)
        kwargs = {k: v.accept(self) for k, v in node.bound_args.kwargs.items()}
        bound_args = node.merge_method.get_signature().bind(*args, **kwargs)
        res = MergeRecipeNode(
            node.merge_method,
            bound_args,
            node.model_config,
            node.merge_space,
        )
        # if self.check_finite and node.merge_method == sd_mecha.extensions.merge_methods.resolve("convert_singleton"):
        #     res = sd_mecha.omit_non_finite(res, self.check_finite_mandatory, self.check_finite_optional)
        self.converted_nodes[node] = res
        return res



import abc
import contextlib
import dataclasses
import threading
from collections import defaultdict
import fuzzywuzzy.process
import inspect
import pathlib
import torch
import typing
import textwrap
import ast
from sd_mecha.recipe_nodes import ClosedModelRecipeNode, RecipeNode, MergeRecipeNode, LiteralRecipeNode, NonDictLiteralValue, RecipeNodeOrValue, PythonLiteralValue
from . import merge_spaces, model_configs
from .merge_spaces import MergeSpace, MergeSpaceSymbol, AnyMergeSpace
from .model_configs import ModelConfig
from types import SimpleNamespace
from typing import Optional, Callable, Dict, Tuple, List, Iterable, Any, Generic, TypeVar, Mapping, Sequence
from ..keys_map import KeyMapBuilder, KeyMap, KeyRelation
from ..typing_ import is_subclass, is_instance


T = TypeVar('T', *typing.get_args(NonDictLiteralValue))


class StateDict(Mapping[str, T], Generic[T], abc.ABC):
    @property
    @abc.abstractmethod
    def model_config(self) -> ModelConfig:
        pass

    @abc.abstractmethod
    def keys(self) -> Iterable[str]:
        pass

    @abc.abstractmethod
    def __contains__(self, item):
        pass


@dataclasses.dataclass
class ParameterData:
    interface: type | TypeVar
    merge_space: Optional[MergeSpace | AnyMergeSpace]
    model_config: Optional[ModelConfig]


@dataclasses.dataclass
class ParameterType:
    data: ParameterData


def Parameter(
    interface: type[NonDictLiteralValue | StateDict[NonDictLiteralValue]] | TypeVar,
    merge_space: Optional[MergeSpace | str | Iterable[MergeSpace | str] | MergeSpaceSymbol] = None,
    model_config: Optional[ModelConfig | str] = None,
) -> type[Any]:
    """
    Describe a parameter to a merge method with its type, optional merge space, and optional model config.

    `Parameter` is used in function type hints to specify how `@merge_method` should handle each
    argument. For example, `Parameter(Tensor, "weight")` means "this argument is a tensor
    in weight space."

    Args:
        interface (type):
            The Python or Torch type of this parameter.
        merge_space (str or Iterable[str], optional):
            Which merge space(s) are valid for this parameter (e.g., "weight", "delta").
        model_config (str or ModelConfig, optional):
            A specific model config or config identifier if it must match a certain architecture.

    Returns:
        A special type annotation object used by `@merge_method` to interpret function arguments.
    """
    if type(None) in (typing.get_args(interface) or ()):
        interface = typing.get_args(interface)[0]

    if not isinstance(interface, TypeVar):
        supported_types = [StateDict] + list(T.__constraints__)
        if not any(issubclass(typing.get_origin(interface) or interface, supported_type) for supported_type in supported_types):
            raise TypeError(f"type {interface} should be one of {', '.join(map(lambda x: x.__name__, supported_types))}")

    if isinstance(merge_space, (str, MergeSpace)):
        merge_space = (merge_space,)

    if isinstance(merge_space, Iterable):
        merge_space = {
            merge_spaces.resolve(m) if isinstance(m, str) else m
            for m in merge_space
        }

    if isinstance(model_config, str):
        model_config = model_configs.resolve(model_config)

    if getattr(model_config, "identifier", None) == "structural":
        raise ValueError("merge methods cannot convert 'structural' model configs")

    return type(Parameter.__name__, (ParameterType,), {
        "data": ParameterData(interface, merge_space, model_config)
    })


@dataclasses.dataclass
class ReturnType:
    data: ParameterData


def Return(
    interface: type[NonDictLiteralValue | StateDict[NonDictLiteralValue]] | TypeVar,
    merge_space: Optional[MergeSpace | str | MergeSpaceSymbol] = None,
    model_config: Optional[ModelConfig | str] = None,
) -> type[Any]:
    """
    Describe the return type of a merge method, optionally including its merge space and model config.

    Args:
        interface (type):
            The Python or Torch type (e.g., `torch.Tensor`) returned by the merge method.
        merge_space (MergeSpace, str or MergeSpaceSymbol, optional):
            The single merge space valid for the return, or a symbol that depends on the input spaces.
        model_config (ModelConfig or str, optional):
            The model config that the returned tensor or dictionary should belong to.

    Returns:
        A type annotation object used by `@merge_method` for the return signature.
    """
    if not isinstance(interface, TypeVar):
        supported_types = [StateDict] + list(T.__constraints__)
        if not any(issubclass(typing.get_origin(interface) or interface, supported_type) for supported_type in supported_types):
            raise TypeError(f"type {interface} should be one of {', '.join(map(lambda x: x.__name__, supported_types))}")

    if isinstance(merge_space, str):
        merge_space = merge_spaces.resolve(merge_space)

    if isinstance(model_config, str):
        model_config = model_configs.resolve(model_config)

    if getattr(model_config, "identifier", None) == "structural":
        raise ValueError("merge methods cannot convert 'structural' model configs")

    return type(Return.__name__, (ReturnType,), {
        "data": ParameterData(interface, merge_space, model_config)
    })


P = TypeVar('P')
@dataclasses.dataclass
class FunctionArgs(Generic[P]):
    args: List[P]
    vararg: P | SimpleNamespace  # using SimpleNamespace as "empty" because P can be Optional
    kwargs: Dict[str, P]

    def as_dict(self, args_count=None) -> Dict[int | str, P]:
        varargs_count = self._get_varargs_count(args_count)
        args_dict = {i: v for i, v in enumerate(self.args)}
        vararg_dict = {
            i: self.vararg
            for i in range(len(args_dict), len(args_dict) + varargs_count)
        }
        return args_dict | vararg_dict | self.kwargs

    def args_varargs(self, args_count=None) -> List[P]:
        varargs_count = self._get_varargs_count(args_count)
        varargs = [self.vararg]*varargs_count
        return self.args + varargs

    def _get_varargs_count(self, args_count):
        n_args = len(self.args)
        if args_count is None:
            args_count = n_args + 1

        return max(0, args_count - n_args) * int(self.has_varargs())

    def has_varargs(self):
        return self.vararg != FunctionArgs.EMPTY_VARARGS


FunctionArgs.EMPTY_VARARGS = SimpleNamespace()


class MergeMethod:
    def __init__(self, fn_or_cls: Callable | type, identifier: str, default_merge_space: MergeSpaceSymbol, interface: Optional["MergeMethodInterface"], reuse_outputs: bool):
        self.__wrapped__ = fn_or_cls
        self.wrapped_is_class = inspect.isclass(fn_or_cls)
        if self.wrapped_is_class:
            if not hasattr(fn_or_cls, "__call__"):
                raise TypeError("class merge methods must define a __call__ method")
            fn = fn_or_cls.__call__
        else:
            fn = fn_or_cls

        self.default_merge_space = default_merge_space
        self.__interface = interface

        signature = inspect.signature(fn)
        return_annotation = _ensure_return(signature.return_annotation, self.default_merge_space)
        self.__f_signature = signature.replace(
            parameters=[
                p.replace(annotation=_ensure_parameter(p.annotation, p.name, p.default, self.default_merge_space, return_annotation.data.interface) if p.annotation != p.empty else p)
                for p in signature.parameters.values()
            ],
            return_annotation=return_annotation
        )

        if self.wrapped_is_class:
            if not isinstance(inspect.getattr_static(fn, "__call__"), (classmethod, staticmethod)):
                self_arg = next(iter(p.name for p in self.__f_signature.parameters.values()))
                self.__f_signature = self.__f_signature.replace(parameters=[v for v in self.__f_signature.parameters.values() if v.name != self_arg])

        self.identifier = identifier
        self.has_varkwargs = any(p.kind == p.VAR_KEYWORD for p in self.__f_signature.parameters.values())
        self.reuse_outputs = reuse_outputs

        self.__validate()

    def __validate(self):
        names = self.get_param_names().as_dict()
        params = self.get_params()  # validates param type annotations
        defaults = self.get_default_args()
        input_merge_spaces = self.get_input_merge_spaces()
        input_configs = self.get_input_configs()

        for param_idx in params.as_dict():
            is_default_arg = (
                isinstance(param_idx, int) and
                len(params.args) - len(defaults.args) <= param_idx < len(params.args)
            )
            is_default_kwarg = isinstance(param_idx, str) and self.__f_signature.parameters[param_idx].default != self.__f_signature.empty
            param_name = names[param_idx]
            if is_default_arg or is_default_kwarg:
                param_merge_space = input_merge_spaces.as_dict()[param_idx]
                if param_merge_space != {merge_spaces.resolve("param")}:
                    raise TypeError(f"The merge space for '{param_name}' should be 'param' since it has a default value.")

        return_data = self.__f_signature.return_annotation.data

        if isinstance(return_data.merge_space, MergeSpaceSymbol):
            if not any(p.merge_space == return_data.merge_space for p in params.as_dict().values()):
                raise RuntimeError("When using a merge space symbol as output, it must also be used by at least one input parameter.")

        configs_involved = (set(getattr(config, "identifier", None) for config in input_configs.as_dict().values()) | {getattr(return_data.model_config, "identifier", None)}).difference({None})
        is_conversion_implicitly = len(configs_involved) > 1
        is_return_dict = is_subclass(return_data.interface, StateDict)
        is_map_keys_defined = self.wrapped_is_class and isinstance(inspect.getattr_static(self.__wrapped__, "map_keys", None), (staticmethod, classmethod))
        if self.__interface is None:
            if (is_conversion_implicitly or is_return_dict) and not is_map_keys_defined:
                raise RuntimeError("A merge method that converts configs must be a class merge method and define a static member 'map_keys(builder)'")
        else:
            if is_map_keys_defined:
                raise RuntimeError("A merge method interface cannot define 'map_keys'.")

    def instantiate(self):
        if self.wrapped_is_class:
            return self.__wrapped__()
        return None

    def key_map(self, args_configs: Sequence[ModelConfig], kwargs_configs: Mapping[str, ModelConfig], return_config: ModelConfig) -> KeyMap:
        input_configs = self.__f_signature.bind(*args_configs, **kwargs_configs).arguments
        input_configs = {
            p.name: input_configs.get(p.name) if input_configs.get(p.name) is not None else p.annotation.data.model_config if p.annotation.data.model_config is not None else return_config
            for p in self.__f_signature.parameters.values()
            if p.kind != p.VAR_KEYWORD
        }
        if self.wrapped_is_class and hasattr(self.__wrapped__, "map_keys"):
            builder = KeyMapBuilder(input_configs, return_config)
            self.__wrapped__.map_keys(builder)
            res = builder.build()
        else:
            res = KeyMap({
                (key,): KeyRelation(
                    (key,),
                    {input_name: (key,) for input_name in input_configs},
                )
                for key in return_config.keys()
            })
        return res

    def __eq__(self, other):
        return isinstance(other, MergeMethod) and self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __repr__(self):
        return f"<merge method '{self.identifier}'>"

    def merge_key(
        self,
        input_args: Sequence[NonDictLiteralValue | StateDict[NonDictLiteralValue]],
        input_kwargs: Mapping[str, NonDictLiteralValue | StateDict[NonDictLiteralValue]],
        key: str,
        key_relation: KeyRelation,
        cache: Optional[dict],
        context: Optional[Any],
        output_reused: bool,
    ) -> NonDictLiteralValue | Mapping[str, NonDictLiteralValue]:
        args, kwargs = self.__get_args_kwargs(input_args, input_kwargs, key, key_relation, cache, output_reused)
        fn = self.__wrapped__
        if self.wrapped_is_class:
            assert context is not None, f"class merge method {self.identifier} received self=None"
            return context(*args, **kwargs)
        return fn(*args, **kwargs)

    def __get_args_kwargs(
        self,
        input_args: Sequence[NonDictLiteralValue | StateDict[NonDictLiteralValue]],
        input_kwargs: Mapping[str, float],
        key: str,
        key_relation: KeyRelation,
        cache: Optional[dict],
        output_reused: bool,
    ) -> Tuple[Sequence[NonDictLiteralValue | StateDict[NonDictLiteralValue]], Mapping]:
        if self.has_varkwargs:
            input_kwargs |= {
                "key": key,
                "key_relation": key_relation,
                "cache": cache,
                "output_reused": output_reused,
            }
        return input_args, input_kwargs

    def __call__(self, *args, **kwargs) -> MergeRecipeNode:
        bound_args = self.__f_signature.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return self.create_recipe(bound_args)

    def create_recipe(self, bound_args: inspect.BoundArguments) -> MergeRecipeNode:
        types = self.get_input_types()
        args = tuple(
            value_to_node(arg, arg_type)
            for arg, arg_type in zip(bound_args.args, types.args_varargs(len(bound_args.args)))
        )
        kwargs = {
            k: value_to_node(arg, types.kwargs[k])
            for k, arg in bound_args.kwargs.items()
        }

        if self.__interface is not None:
            return self.__interface.dispatch(*args, **kwargs)

        return MergeRecipeNode(self, self.__f_signature.bind(*args, **kwargs))

    def get_signature(self):
        return self.__f_signature

    def get_return_type(self) -> type:
        return self.__f_signature.return_annotation

    def get_input_types(self) -> FunctionArgs[type]:
        params = self.get_params()
        arg_types = [
            param.interface
            for param in params.args
        ]
        vararg_type = params.vararg.interface if params.has_varargs() else FunctionArgs.EMPTY_VARARGS
        kwarg_types = {
            k: param.interface
            for k, param in params.kwargs.items()
        }
        return FunctionArgs(arg_types, vararg_type, kwarg_types)

    def get_input_merge_spaces(self) -> FunctionArgs[AnyMergeSpace]:
        params = self.get_params()
        names = self.get_param_names()
        defaults = self.get_default_args()

        def get_merge_space_or_default(param, has_default):
            merge_space = param.merge_space
            if merge_space is None:
                if has_default or is_subclass(param.interface, PythonLiteralValue):
                    merge_space = {merge_spaces.resolve("param")}
                else:
                    merge_space = self.default_merge_space
            return merge_space

        args_merge_spaces = []
        for i in range(len(names.args)):
            param = params.args[i]
            merge_space = get_merge_space_or_default(param, i >= len(names.args) - len(defaults.args))
            args_merge_spaces.append(merge_space)

        varargs_merge_space = FunctionArgs.EMPTY_VARARGS
        if params.has_varargs():
            varargs_merge_space = params.vararg.merge_space
            if varargs_merge_space is None:
                varargs_merge_space = self.default_merge_space

        kwargs_merge_spaces = {}
        for name in names.kwargs:
            param = params.kwargs[name]
            merge_space = get_merge_space_or_default(param, name in defaults.kwargs)
            kwargs_merge_spaces[name] = merge_space

        return FunctionArgs(args_merge_spaces, varargs_merge_space, kwargs_merge_spaces)

    def get_input_configs(self) -> FunctionArgs[Optional[ModelConfig]]:
        params = self.get_params()
        return FunctionArgs(
            [arg.model_config for arg in params.args],
            params.vararg.model_config if params.has_varargs() else FunctionArgs.EMPTY_VARARGS,
            {k: v.model_config for k, v in params.kwargs.items()},
        )

    def get_param_names(self) -> FunctionArgs[str]:
        return FunctionArgs(
            [p.name for p in self.__f_signature.parameters.values() if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD)],
            tuple((*(p.name for p in self.__f_signature.parameters.values() if p.kind == p.VAR_POSITIONAL), FunctionArgs.EMPTY_VARARGS))[0],
            {p.name: p.name for p in self.__f_signature.parameters.values() if p.kind == p.KEYWORD_ONLY},
        )

    def get_default_args(self) -> FunctionArgs[Any]:
        return FunctionArgs(
            [p.default for p in self.__f_signature.parameters.values() if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD) and p.default != p.empty],
            tuple((*(p for p in self.__f_signature.parameters.values() if p.kind == p.VAR_POSITIONAL), FunctionArgs.EMPTY_VARARGS))[0],
            {p.name: p.default for p in self.__f_signature.parameters.values() if p.kind == p.KEYWORD_ONLY and p.default != p.empty},
        )

    def get_params(self) -> FunctionArgs[ParameterData]:
        names = self.get_param_names()

        return FunctionArgs(
            [self.__f_signature.parameters[k].annotation.data for k in names.args],
            self.__f_signature.parameters[names.vararg].annotation.data if names.has_varargs() else FunctionArgs.EMPTY_VARARGS,
            {k: self.__f_signature.parameters[k].annotation.data for k in names.kwargs},
        )

    def get_identifier(self) -> str:
        return self.identifier


class MergeMethodInterface:
    def __init__(self, identifier: str, fn: Callable, default_merge_space: MergeSpaceSymbol):
        self.identifier = identifier
        self.candidates = []

        self.default_merge_space = default_merge_space

        signature = inspect.signature(fn)
        for param in signature.parameters.values():
            if param.kind == inspect.Parameter.KEYWORD_ONLY:
                raise RuntimeError(f"Keyword-only parameter '{param.name}' is not allowed in a merge strategy.")
        return_annotation = _ensure_return(signature.return_annotation, self.default_merge_space)
        self.signature = signature.replace(
            parameters=[
                p.replace(annotation=_ensure_parameter(p.annotation, p.name, p.default, self.default_merge_space, return_annotation.data.interface))
                for p in signature.parameters.values()
                if p.kind != inspect.Parameter.VAR_KEYWORD
            ],
            return_annotation=return_annotation,
        )

    def register_implementation(self, candidate: MergeMethod):
        candidate_signature = candidate.get_signature()
        new_parameters = []

        for (contract_name, contract_param), (candidate_name, candidate_param) in zip(
            self.signature.parameters.items(), candidate_signature.parameters.items(),
        ):
            contract_param: inspect.Parameter
            candidate_param: inspect.Parameter
            if candidate_param.kind != contract_param.kind:
                raise RuntimeError(f"Expected parameter '{candidate_name}' to be {contract_param.kind} but is {candidate_param.kind}.")
            if candidate_name != contract_name:
                raise RuntimeError(f"Expected parameter '{candidate_name}' to be named '{contract_name}'.")

            candidate_data = candidate_param.annotation.data
            contract_data = contract_param.annotation.data
            if candidate_data.interface != contract_data.interface:
                raise TypeError(f"Expected parameter '{candidate_name}' to have type {contract_data.interface} but got {candidate_data.interface}.")
            if contract_data.merge_space is not None and candidate_data.merge_space != contract_data.merge_space:
                raise TypeError(f"Expected parameter '{candidate_name}' to use merge space(s) {contract_data.merge_space} but got {candidate_data.merge_space}.")
            if contract_data.model_config is not None and candidate_data.model_config != contract_data.model_config:
                raise TypeError(f"Expected parameter '{candidate_name}' to use model config {contract_data.model_config} but got {candidate_data.model_config}.")
            if contract_param.default == inspect.Parameter.empty and candidate_param.default != inspect.Parameter.empty:
                raise TypeError(f"Expected parameter '{candidate_name}' to have no default value.")

            new_parameters.append(candidate_param.replace(
                default=candidate_param.default if candidate_param.default != inspect.Parameter.empty else contract_param.default,
            ))

        candidate_data = candidate_signature.return_annotation.data
        contract_data = self.signature.return_annotation.data
        if not _same_state_dict_interface(candidate_data.interface, contract_data.interface):
            raise TypeError(f"Expected return type {contract_data.interface} but got {candidate_data.interface}.")
        if contract_data.merge_space is not None and candidate_data.merge_space != contract_data.merge_space:
            raise TypeError(f"Expected return merge space {contract_data.merge_space} but got {candidate_data.merge_space}.")
        if contract_data.model_config is not None and candidate_data.model_config != contract_data.model_config:
            raise TypeError(f"Expected return model config {contract_data.model_config} but got {candidate_data.model_config}.")

        candidate_signature = candidate_signature.replace(parameters=new_parameters)
        self.candidates.append((candidate, candidate_signature))

    def dispatch(self, *args, **kwargs):
        argument_graphs = {}
        failures: list[tuple[str, str]] = []

        with contextlib.ExitStack() as stack:
            for candidate, candidate_signature in self.candidates:
                try:
                    bound_args = self._try_candidate(
                        candidate,
                        candidate_signature,
                        argument_graphs,
                        stack,
                        args,
                        kwargs,
                    )
                    return candidate.create_recipe(bound_args)
                except CandidateDispatchError as e:
                    failures.append((candidate.identifier, str(e)))
                except Exception as e:
                    failures.append((candidate.identifier, f"unexpected error: {e}"))
        if failures:
            lines = [
                f"No candidate matched the given arguments for interface '{self.identifier}'.",
                "Attempted candidates:",
            ]
            lines.extend(f"  - {candidate_id}: {reason}" for candidate_id, reason in failures)
            raise TypeError("\n".join(lines))

        raise TypeError(f"No candidate matched the given arguments: {self.identifier}(*{args}, **{kwargs})")

    def _try_candidate(
        self,
        candidate: MergeMethod,
        candidate_signature: inspect.Signature,
        argument_graphs: dict,
        stack: contextlib.ExitStack,
        args,
        kwargs,
    ) -> inspect.BoundArguments:
        from sd_mecha import open_graph

        try:
            bound_args: inspect.BoundArguments = candidate_signature.bind(*args, **kwargs)
        except TypeError as e:
            raise CandidateDispatchError(candidate.identifier, f"signature mismatch: {e}") from e

        for parameter_name, argument_node in bound_args.arguments.copy().items():
            contract_data = candidate_signature.parameters[parameter_name].annotation.data
            contract_model_config = contract_data.model_config
            contract_merge_spaces = contract_data.merge_space

            try:
                if argument_node not in argument_graphs:
                    argument_graph = argument_graphs[argument_node] = stack.enter_context(
                        open_graph(argument_node, root_only=True)
                    )
                else:
                    argument_graph = argument_graphs[argument_node]

                argument_candidates = argument_graph.root_candidates(
                    model_config=contract_model_config,
                    merge_space_preference=(contract_merge_spaces,),
                )
            except Exception as e:
                raise CandidateDispatchError(
                    candidate.identifier,
                    f"parameter '{parameter_name}': failed to analyze argument graph: {e}"
                ) from e

            mc_satisfied = bool(argument_candidates.model_config)
            ms_satisfied = contract_merge_spaces is None or (
                    argument_candidates.merge_space is not None and
                    any(ms in contract_merge_spaces for ms in argument_candidates.merge_space)
            )

            if not mc_satisfied or not ms_satisfied:
                details = []
                if contract_model_config is not None and not mc_satisfied:
                    details.append(f"model_config requirement not satisfied ({contract_model_config.identifier})")
                if contract_merge_spaces is not None and not ms_satisfied:
                    want = sorted(ms.identifier for ms in contract_merge_spaces)
                    got = (
                        sorted(ms.identifier for ms in argument_candidates.merge_space)
                        if argument_candidates.merge_space is not None else
                        None
                    )
                    details.append(f"merge_space requirement not satisfied (wanted one of {want}, got {got})")

                raise CandidateDispatchError(
                    candidate.identifier,
                    f"parameter '{parameter_name}': " + "; ".join(details)
                )

        bound_args.apply_defaults()
        return bound_args


class CandidateDispatchError(TypeError):
    def __init__(self, candidate_identifier: str, message: str):
        super().__init__(message)
        self.candidate_identifier = candidate_identifier


def _same_state_dict_interface(i1, i2):
    if typing.get_origin(i1) == StateDict:
        i1 = typing.get_args(i1)[0]
    if typing.get_origin(i2) == StateDict:
        i2 = typing.get_args(i2)[0]
    return i1 == i2


def _ensure_parameter(hint: type, param_name: str, default: Any, default_merge_space: MergeSpaceSymbol, return_interface: type):
    # remove outer `| None`
    hint_args = [arg for arg in (typing.get_args(hint) or ()) if arg is not type(None)]
    if hint_args:
        hint = hint_args[0]

    if hint is Parameter:
        raise TypeError(f"the type of parameter '{param_name}' should be `sd_mecha.Parameter(...)`, not `sd_mecha.Parameter` (note the lack of parentheses)")

    if not inspect.isclass(hint) or not issubclass(hint, ParameterType):
        hint = Parameter(hint)

    if hint.data.merge_space is None:
        interface = hint.data.interface
        if is_subclass(hint.data.interface, StateDict):
            interface = (typing.get_args(interface) or (T,))[0]

        if is_subclass(return_interface, StateDict):
            return_interface = (typing.get_args(return_interface) or (T,))[0]

        if default is inspect.Parameter.empty and is_subclass(interface, return_interface):
            hint.data.merge_space = default_merge_space
        else:
            hint.data.merge_space = {"param"}

    return hint


def _ensure_return(hint: type, default_merge_space: MergeSpaceSymbol):
    if hint is Return:
        raise TypeError(f"the return type should be 'sd_mecha.Return(...)', not 'sd_mecha.Return' (note the lack of parentheses)")

    if not inspect.isclass(hint) or not issubclass(hint, ReturnType):
        hint = Return(hint)

    if hint.data.merge_space is None:
        hint.data.merge_space = default_merge_space

    return hint


F = TypeVar("F", bound=Callable | type)


def merge_method(
    fn: Optional[F] = None, *,
    identifier: Optional[str] = None,
    register: bool = True,
    is_conversion: bool = False,
    implements: Optional[str | MergeMethod] = None,
    is_interface: bool = False,
    reuse_outputs: bool = True,
) -> MergeMethod | Callable[[F], MergeMethod]:
    """
    Decorator to define a custom merge method.

    This converts the decorated function into a `MergeMethod` object that can be used to create
    recipe graphs. Use sd_mecha.Parameter(...) as the type hint of parameters to add constraints to the inputs
    (i.e. merge space, model config).

    Args:
        fn (callable, optional):
            The function to convert to a merge method object.
        identifier (str, optional):
            An explicit name to register this method under. By default, uses `fn.__name__`.
        register (bool):
            If True (default), registers the merge method globally so it can be accessed by `merge_methods.resolve()`.
        is_conversion (bool):
            If True, marks this merge method as a config-conversion function. That means `convert()`
            will consider it as an implicit transition when converting between different model configs.
        implements (str | MergeMethod):
            The interface to implement. The signature of the merge method must match that of the interface.
        is_interface (bool):
            If True, marks this merge method can be overloaded with multiple implementations.
            The appropriate candidate implementation will be resolved during recipe node creation.
        reuse_outputs (bool):
            If True, marks this merge method as computationally heavy.

    Returns:
        A `MergeMethod` object or a decorator returning such an object.

    Raises:
        ValueError: another merge method with this identifier was already registered.
        ValueError: register is False, but is_conversion is True or dispatcher is not None.
    """
    if fn is None:
        return lambda fn: _merge_method_impl(fn, identifier=identifier, register=register, is_conversion=is_conversion, implements=implements, is_interface=is_interface, reuse_outputs=reuse_outputs)
    return _merge_method_impl(fn, identifier=identifier, register=register, is_conversion=is_conversion, implements=implements, is_interface=is_interface, reuse_outputs=reuse_outputs)


def _merge_method_impl(
    fn: F, *,
    identifier: Optional[str],
    register: bool,
    is_conversion: bool,
    implements: Optional[str | MergeMethod],
    is_interface: bool,
    reuse_outputs: bool,
) -> MergeMethod:
    global _module_state

    if identifier is None:
        identifier = fn.__name__

    if isinstance(implements, MergeMethod):
        implements = implements.identifier

    if not register:
        if is_conversion:
            raise ValueError("A conversion merge method must be registered.")
        if implements is not None:
            raise ValueError("A merge method overload must be registered.")
        if is_interface:
            raise ValueError("A merge method interface must be registered.")

    if is_interface and implements is not None:
        raise ValueError("An merge method interface cannot overload a merge method interface.")

    if implements is not None and implements not in _module_state.interfaces_registry:
        raise ValueError(f"The provided merge method interface {implements} is not an interface.")

    with _module_state.registry_lock:
        if register:
            if identifier in _module_state.merge_methods_registry:
                raise ValueError(f"Another merge method named {identifier} is already registered.")

        module_state_copy = _module_state.copy()

        if implements is not None:
            default_merge_space = module_state_copy.merge_methods_registry[implements].default_merge_space
        else:
            default_merge_space = MergeSpaceSymbol(*merge_spaces.get_all())

        if is_interface:
            _register_interface(fn, identifier, default_merge_space, module_state_copy)

        fn_object = MergeMethod(fn, identifier, default_merge_space, module_state_copy.interfaces_registry.get(identifier), reuse_outputs)
        module_state_copy.merge_methods_registry[identifier] = fn_object

        if is_conversion:
            _register_config_converter(fn_object, module_state_copy)
        if implements is not None:
            module_state_copy.interfaces_registry[implements].register_implementation(fn_object)

        _module_state = module_state_copy

    return fn_object


@dataclasses.dataclass
class ModuleState:
    registry_lock: threading.Lock = dataclasses.field(default_factory=threading.Lock)
    merge_methods_registry: Dict[str, MergeMethod] = dataclasses.field(default_factory=dict)
    conversion_registry: Dict[str, MergeMethod] = dataclasses.field(default_factory=dict)
    converter_paths: Dict[str, List[Tuple[str, MergeMethod]]] = dataclasses.field(default_factory=lambda: defaultdict(list))
    interfaces_registry: Dict[str, MergeMethodInterface] = dataclasses.field(default_factory=dict)

    def copy(self):
        return ModuleState(
            self.registry_lock,
            self.merge_methods_registry.copy(),
            self.conversion_registry.copy(),
            defaultdict(list, ((k, v.copy()) for k, v in self.converter_paths.items())),
            self.interfaces_registry.copy(),
        )


_module_state = ModuleState()


def resolve(identifier: str) -> MergeMethod:
    try:
        return _module_state.merge_methods_registry[identifier]
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _module_state.merge_methods_registry.keys())[0]
        raise KeyError(f"unknown merge method: {e}. Nearest match is '{suggestion}'")


def get_all() -> List[MergeMethod]:
    return list(_module_state.merge_methods_registry.values())


def get_all_converters() -> List[MergeMethod]:
    return list(_module_state.conversion_registry.values())


def get_converter_paths() -> Dict[str, List[Tuple[str, MergeMethod]]]:
    return _module_state.converter_paths.copy()


def _register_config_converter(converter: MergeMethod, module_state: ModuleState):
    validate_config_converter(converter)
    input_configs = converter.get_input_configs()
    return_config = converter.get_signature().return_annotation.data.model_config
    src_config = input_configs.args[0].identifier
    tgt_config = return_config.identifier if return_config is not None else None

    module_state.conversion_registry[converter.identifier] = converter
    module_state.converter_paths[src_config].append((tgt_config, converter))


def _register_interface(
    fn: Callable,
    identifier: Optional[str],
    default_merge_space: MergeSpaceSymbol,
    module_state: ModuleState,
) -> MergeMethodInterface:
    if identifier is None:
        identifier = fn.__name__
    fn_object = MergeMethodInterface(identifier, fn, default_merge_space)

    if identifier in module_state.interfaces_registry:
        raise KeyError(f"Another merge method interface named {identifier} is already registered.")

    if not _is_empty_body_fn(fn):
        raise ValueError("register() must be applied to an empty function (the body must be either nothing, `pass`, `...`, `return` or `return None`. docstrings are allowed)")

    module_state.interfaces_registry[identifier] = fn_object
    return fn_object


def validate_config_converter(merge_method: MergeMethod):
    params = merge_method.get_param_names()
    args_varargs = params.args if params.args else params.args_varargs()
    assert len(args_varargs) == 1, f"the merge method should be able to take exactly 1 positional argument"
    configs = merge_method.get_input_configs()
    input_config = configs.args[0] if configs.args else configs.args_varargs()[0]
    assert input_config is not None, f"the input model config is missing. It should be declared in the type of `{args_varargs[0]}`"
    return merge_method


def value_to_node(node_or_value: RecipeNodeOrValue, expected_type: type = None) -> RecipeNode:
    """
    Convert a literal or path into a `RecipeNode`, if it isn't one already.

    This helper is primarily used internally when constructing recipe graphs. For instance,
    if you pass a float to a merge method where a tensor is expected, `value_to_node` wraps
    that float in a node that will automatically convert it into a tensor when broadcasting
    over state dict keys.

    Args:
        node_or_value:
            The input that should become a `RecipeNode`.
        expected_type (type, optional):
            The target type that the value should be converted to.

    Returns:
        A `RecipeNode` instance, if conversion is successful.
    """
    if isinstance(node_or_value, RecipeNode):
        return node_or_value

    if not is_instance(node_or_value, RecipeNodeOrValue):
        raise TypeError(f"type of 'node_or_value' should be one of {typing.get_args(RecipeNodeOrValue)}, not {type(node_or_value)}")

    if expected_type is None:
        if isinstance(node_or_value, Mapping):
            expected_type = type(next(iter(node_or_value.values())))
        else:
            expected_type = type(node_or_value)

        if issubclass(expected_type, pathlib.Path):
            expected_type = torch.Tensor

    numeric = int | float

    # verify dict value type consistency
    if isinstance(node_or_value, Mapping) and node_or_value:
        actual_type = type(next(iter(node_or_value.values())))
        if is_subclass(actual_type, numeric):
            actual_type = numeric
        if is_subclass(actual_type, RecipeNode):
            actual_type = RecipeNode
        if not is_subclass(actual_type, NonDictLiteralValue | RecipeNode):
            raise TypeError(f"unsupported type found in input dict: {actual_type} (supported types are {typing.get_args(NonDictLiteralValue)})")
        if not all(isinstance(v, actual_type) for v in node_or_value.values()):
            bad_type = next(iter(type(v) for v in node_or_value.values() if not isinstance(v, actual_type)))
            raise TypeError(f"inconsistent types found in input dict: {actual_type} and {bad_type}")
        if is_subclass(actual_type, RecipeNode):
            first_node = next(iter(node_or_value.values()))
            if not all(r.model_config == first_node.model_config for r in node_or_value.values()):
                raise RuntimeError("inconsistent model configs found in recipe dict")
            if not all(r.merge_space == first_node.merge_space for r in node_or_value.values()):
                raise RuntimeError("inconsistent output merge space found in recipe dict")

    try:
        if issubclass(typing.get_origin(expected_type) or expected_type, StateDict):
            expected_type = next(iter(typing.get_args(expected_type) or (T,)))
    except TypeError:
        pass

    if isinstance(node_or_value, NonDictLiteralValue):
        res = LiteralRecipeNode({"key": node_or_value}, model_config="singleton-mecha")
        res = _module_state.merge_methods_registry["convert_singleton"](res)
        return res

    if isinstance(node_or_value, dict):
        return LiteralRecipeNode(node_or_value)

    if isinstance(node_or_value, pathlib.Path) and (isinstance(expected_type, TypeVar) or issubclass(expected_type, torch.Tensor)):
        try:
            return ClosedModelRecipeNode(node_or_value)
        except TypeError as e:
            base_error_message = f"No implicit conversion exists from {type(node_or_value)} to Dict[Tensor]"
            if isinstance(node_or_value, str):
                raise TypeError(f"{base_error_message}. Consider using pathlib.Path instead if a model path was intended") from e
            if not isinstance(node_or_value, pathlib.Path):
                raise TypeError(base_error_message) from e

    raise TypeError(f"No implicit conversion exists from {type(node_or_value)} to {expected_type}")


def _is_empty_body_fn(func) -> bool:
    try:
        src = inspect.getsource(func)
    except (OSError, TypeError):  # no source available (builtins, C-ext, interactive, etc.)
        return False

    src = textwrap.dedent(src)
    node = ast.parse(src)

    # Find the first function/async function node in that source block
    fn = next(
        (n for n in node.body if isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef))),
        None
    )
    if fn is None:
        return False

    # Ignore leading docstring if present
    body = fn.body[:]
    if (
        body
        and isinstance(body[0], ast.Expr)
        and isinstance(getattr(body[0], "value", None), ast.Constant)
        and isinstance(body[0].value.value, str)
    ):
        body = body[1:]

    # Empty after docstring => "empty"
    if not body:
        return True

    # Exactly one statement: `pass`, `ellipsis`, `return` or `return None`
    if len(body) == 1:
        stmt = body[0]
        if isinstance(stmt, ast.Pass):
            return True
        if (
            isinstance(stmt, ast.Expr)
            and isinstance(getattr(stmt, "value", None), ast.Constant)
            and stmt.value.value is Ellipsis
        ):
            return True
        if (
            isinstance(stmt, ast.Return)
            and (
                stmt.value is None or
                isinstance(stmt.value, ast.Constant) and stmt.value.value is None
            )
        ):
            return True

    return False



import dataclasses
from collections import defaultdict, deque
from types import MappingProxyType
from typing import DefaultDict, Dict, List, Set, Tuple, Mapping
from sd_mecha.streaming import TensorMetadata


@dataclasses.dataclass(frozen=True, slots=True)
class PermSlice:
    axis: int
    offset: int = 0


@dataclasses.dataclass(frozen=True, slots=True)
class Permutation:
    width: int
    slices: Mapping[str, PermSlice | Tuple[PermSlice, ...]]

    def keys(self):
        return self.slices.keys()


@dataclasses.dataclass(frozen=True, slots=True)
class PermNode:
    """One window for one key inside one permutation."""
    key: str
    axis: int
    offset: int


@dataclasses.dataclass(frozen=True, slots=True)
class PermClosure:
    """
    Invocation unit produced by PermGraph.closures().

    Existing fields (kept):
      - perms:         permutation ids participating in this invocation unit
      - components:    interacting components of perms (solver scheduling)
      - required_keys: keys that must be available as inputs

    Added fields (precomputed to avoid per-call work):
      - pid_to_width:  pid -> width
      - pid_to_nodes:  pid -> tuple[PermNode,...] (expanded slices)
      - key_to_apps:   key -> tuple[(pid, axis, offset), ...] for all perms in this closure
                       (used to apply other-axis perms / final inverse perms)
    """
    perms: Tuple[int, ...]
    components: Tuple[Tuple[int, ...], ...]
    keys: Tuple[str, ...]

    pid_to_width: Mapping[int, int] = dataclasses.field(default_factory=dict)
    pid_to_nodes: Mapping[int, Tuple[PermNode, ...]] = dataclasses.field(default_factory=dict)
    key_to_apps: Mapping[str, Tuple[Tuple[int, int, int], ...]] = dataclasses.field(default_factory=dict)

    @staticmethod
    def build_from_graph(
        *,
        perms: Tuple[int, ...],
        components: Tuple[Tuple[int, ...], ...],
        keys: Tuple[str, ...],
        pid_to_perm: Mapping[int, Permutation],
    ) -> "PermClosure":
        # Expand all (key -> (slice or tuple[slice])) into pid->nodes and key->apps.
        pid_to_width: Dict[int, int] = {}
        pid_to_nodes: Dict[int, Tuple[PermNode, ...]] = {}
        key_apps: Dict[str, List[Tuple[int, int, int]]] = {}

        for pid in perms:
            p = pid_to_perm[pid]
            pid_to_width[pid] = int(p.width)

            nodes: List[PermNode] = []
            for k, sls in p.slices.items():
                if isinstance(sls, PermSlice):
                    sls_iter = (sls,)
                else:
                    sls_iter = tuple(sls)

                for sl in sls_iter:
                    nodes.append(PermNode(k, int(sl.axis), int(sl.offset)))
                    key_apps.setdefault(k, []).append((pid, int(sl.axis), int(sl.offset)))

            pid_to_nodes[pid] = tuple(nodes)

        # Deterministic ordering for applications per key:
        # (axis, offset, pid) so sequential application is stable.
        key_to_apps: Dict[str, Tuple[Tuple[int, int, int], ...]] = {}
        for k, apps in key_apps.items():
            apps.sort(key=lambda t: (t[1], t[2], t[0]))
            key_to_apps[k] = tuple(apps)

        return PermClosure(
            perms=perms,
            components=components,
            keys=keys,
            pid_to_width=MappingProxyType(pid_to_width),
            pid_to_nodes=MappingProxyType(pid_to_nodes),
            key_to_apps=MappingProxyType(key_to_apps),
        )


class PermGraph:
    def __init__(self, md: Mapping[str, TensorMetadata]):
        self.md = md
        self._perms: List[Permutation] = []
        self._by_key: DefaultDict[str, List[int]] = defaultdict(list)

    def add(self, p: Permutation) -> int:
        """Register one permutation; returns its integer id (pid)."""
        pid = len(self._perms)
        _validate_perm_disjoint(self.md, p)

        self._perms.append(p)
        for k in p.keys():
            self._by_key[k].append(pid)
        return pid

    def keys_with_perms(self) -> Set[str]:
        return set(self._by_key.keys())

    def closures(self) -> List[PermClosure]:
        """
        Return invocation units.

        Closure criterion: permutations are connected if they share ANY key.
        This is the unit you must put on the LHS of b[...] = ...
        """
        # Build perm adjacency for closure connectivity (share any key)
        perm_adj: DefaultDict[int, Set[int]] = defaultdict(set)
        for k, pids in self._by_key.items():
            # connect all perms touching the same key
            for i in range(len(pids)):
                a = pids[i]
                for j in range(i + 1, len(pids)):
                    b = pids[j]
                    perm_adj[a].add(b)
                    perm_adj[b].add(a)

        visited: Set[int] = set()
        out: List[PermClosure] = []

        for start in range(len(self._perms)):
            if start in visited:
                continue

            # BFS over permutations via "share any key" adjacency
            q = deque([start])
            visited.add(start)
            closure_pids: List[int] = [start]
            closure_keys: Set[str] = set(self._perms[start].keys())

            while q:
                cur = q.popleft()
                for nxt in perm_adj.get(cur, ()):
                    if nxt not in visited:
                        visited.add(nxt)
                        q.append(nxt)
                        closure_pids.append(nxt)
                        closure_keys.update(self._perms[nxt].keys())

            # For this invocation unit, compute solver interaction components
            components = self._interaction_components(tuple(closure_pids))

            # required_keys: union of all keys touched by the closure's perms
            keys = tuple(k for k in self.md.keys() if k in closure_keys)

            out.append(PermClosure.build_from_graph(
                perms=tuple(sorted(closure_pids)),
                components=components,
                keys=keys,
                pid_to_perm={pid: self._perms[pid] for pid in closure_pids},
            ))

        return out

    def _interaction_components(self, pids: Tuple[int, ...]) -> Tuple[Tuple[int, ...], ...]:
        """
        Partition pids into connected components where edges exist iff _perms_interact(md, p, q).
        """
        if not pids:
            return tuple()

        # Build interaction adjacency within this closure
        adj: DefaultDict[int, Set[int]] = defaultdict(set)
        for i in range(len(pids)):
            a = pids[i]
            for j in range(i + 1, len(pids)):
                b = pids[j]
                if _perms_interact(self.md, self._perms[a], self._perms[b]):
                    adj[a].add(b)
                    adj[b].add(a)

        remaining: Set[int] = set(pids)
        comps: List[Tuple[int, ...]] = []

        while remaining:
            start = next(iter(remaining))
            q = deque([start])
            remaining.remove(start)
            comp: List[int] = [start]

            while q:
                cur = q.popleft()
                for nxt in adj.get(cur, ()):
                    if nxt in remaining:
                        remaining.remove(nxt)
                        q.append(nxt)
                        comp.append(nxt)

            comp.sort()
            comps.append(tuple(comp))

        comps.sort(key=lambda c: c[0])
        return tuple(comps)


def _perms_interact(md: Mapping[str, TensorMetadata], p: Permutation, q: Permutation) -> bool:
    shared = set(p.keys()) & set(q.keys())
    for k in shared:
        rank = len(md[k].shape)

        ps = p.slices[k]
        qs = q.slices[k]

        if not isinstance(ps, tuple):
            ps = ps,
        if not isinstance(qs, tuple):
            qs = qs,

        for pn in ps:
            pax = _norm_axis(pn.axis, rank)
            p0, p1 = pn.offset, pn.offset + p.width

            for qn in qs:
                qax = _norm_axis(qn.axis, rank)
                if pax != qax:
                    return True
                q0, q1 = qn.offset, qn.offset + q.width
                if _overlap(p0, p1, q0, q1):
                    return True
    return False


def _overlap(a0: int, a1: int, b0: int, b1: int) -> bool:
    # [a0,a1) overlaps [b0,b1)
    return not (a1 <= b0 or b1 <= a0)


def _validate_perm_disjoint(md: Mapping[str, TensorMetadata], p: Permutation) -> None:
    for k, sls in p.slices.items():
        if k not in md:
            raise KeyError(f"unknown key: {k}")
        rank = len(md[k].shape)

        if not isinstance(sls, tuple):
            sls = sls,

        # group by normalized axis
        by_axis: Dict[int, List[int]] = defaultdict(list)
        for sl in sls:
            ax = _norm_axis(sl.axis, rank)
            by_axis[ax].append(sl.offset)

        # validate bounds + disjointness per axis
        for ax, offsets in by_axis.items():
            dim = _axis_size(md, k, ax)
            # bounds
            for off in offsets:
                if off < 0 or off + p.width > dim:
                    raise ValueError(
                        f"{k}: slice [{off}:{off+p.width}] out of bounds for axis {ax} dim {dim}"
                    )
            # disjointness: sort by offset, adjacent must not overlap
            offsets.sort()
            for i in range(len(offsets) - 1):
                a0, a1 = offsets[i], offsets[i] + p.width
                b0, b1 = offsets[i + 1], offsets[i + 1] + p.width
                if _overlap(a0, a1, b0, b1):
                    raise ValueError(
                        f"{k}: overlapping slices on axis {ax}: "
                        f"[{a0}:{a1}] overlaps [{b0}:{b1}] (width={p.width})"
                    )


def _axis_size(md: Mapping[str, TensorMetadata], key: str, axis: int) -> int:
    t = md[key]
    ax = _norm_axis(axis, len(t.shape))
    return int(t.shape[ax])


def _norm_axis(axis: int, rank: int) -> int:
    if axis < 0:
        axis += rank
    if axis < 0 or axis >= rank:
        raise ValueError(f"axis {axis} out of range for rank {rank}")
    return axis



import torch


def convert_n_to_1(kohya_sd, kohya_keys, transpose=False):
    res = [kohya_sd[key] for key in kohya_keys]
    if len(res) > 1:
        res = torch.vstack(res)
    else:
        res = res[0]

    if transpose:
        res = res.T

    return res



import torch
from sd_mecha.extensions.merge_methods import merge_method, StateDict, Parameter, Return
from sd_mecha.extensions import model_configs
from .convert_huggingface_sd_vae_to_original import convert_vae_key, reshape_weight_for_sd


# hf to sd conversion src:
# https://github.com/huggingface/diffusers/blob/main/scripts/convert_diffusers_to_original_stable_diffusion.py


sd1_kohya = model_configs.resolve("sd1-kohya")
sd1_ldm = model_configs.resolve("sd1-ldm")


@merge_method(
    identifier=f"convert_'{sd1_kohya.identifier}'_to_'{sd1_ldm.identifier}'",
    is_conversion=True,
    reuse_outputs=False,
)
class convert_sd1_kohya_to_original:
    @staticmethod
    def map_keys(b):
        for output_key in sd1_ldm.keys():
            needs_reshape = False
            if output_key.startswith("model.diffusion_model."):
                input_keys = convert_unet_key(output_key)
            elif output_key.startswith("cond_stage_model."):
                input_keys = convert_clip_l_key(output_key)
            elif output_key.startswith("first_stage_model."):
                input_keys, needs_reshape = convert_vae_key(output_key)
            else:
                input_keys = output_key
            b[output_key] = b.keys[input_keys] @ needs_reshape

    def __call__(
        self,
        kohya_sd: Parameter(StateDict[torch.Tensor], model_config=sd1_kohya),
        **kwargs,
    ) -> Return(torch.Tensor, model_config=sd1_ldm):
        relation = kwargs["key_relation"]
        needs_reshape = relation.meta

        res = kohya_sd[relation["kohya_sd"][0]]
        if needs_reshape:
            res = reshape_weight_for_sd(res)
        return res


def convert_unet_key(ldm_key: str) -> str:
    kohya_key = '.'.join(ldm_key.split(".")[2:])  # model.diffusion_model.

    for sd_part, hf_part in unet_conversion_map_layer.items():
        kohya_key = kohya_key.replace(sd_part, hf_part)

    if "resnets" in kohya_key:
        for sd_part, hf_part in unet_conversion_map_resnet.items():
            kohya_key = kohya_key.replace(sd_part, hf_part)

    kohya_key = unet_conversion_map.get(kohya_key, kohya_key)

    kohya_key = f"unet.{kohya_key}"
    return kohya_key


unet_conversion_map = {
    # (original: huggingface)
    "time_embed.0.weight": "time_embedding.linear_1.weight",
    "time_embed.0.bias": "time_embedding.linear_1.bias",
    "time_embed.2.weight": "time_embedding.linear_2.weight",
    "time_embed.2.bias": "time_embedding.linear_2.bias",
    "input_blocks.0.0.weight": "conv_in.weight",
    "input_blocks.0.0.bias": "conv_in.bias",
    "out.0.weight": "conv_norm_out.weight",
    "out.0.bias": "conv_norm_out.bias",
    "out.2.weight": "conv_out.weight",
    "out.2.bias": "conv_out.bias",
}


unet_conversion_map_resnet = {
    # (original: huggingface)
    "in_layers.0": "norm1",
    "in_layers.2": "conv1",
    "out_layers.0": "norm2",
    "out_layers.3": "conv2",
    "emb_layers.1": "time_emb_proj",
    "skip_connection": "conv_shortcut",
}


unet_conversion_map_layer = {}
# hardcoded number of downblocks and resnets/attentions...
# would need smarter logic for other networks.
for i in range(4):
    # loop over downblocks/upblocks

    for j in range(2):
        # loop over resnets/attentions for downblocks
        hf_down_res_prefix = f"down_blocks.{i}.resnets.{j}."
        sd_down_res_prefix = f"input_blocks.{3*i + j + 1}.0."
        unet_conversion_map_layer[sd_down_res_prefix] = hf_down_res_prefix

        if i < 3:
            # no attention layers in down_blocks.3
            hf_down_atn_prefix = f"down_blocks.{i}.attentions.{j}."
            sd_down_atn_prefix = f"input_blocks.{3*i + j + 1}.1."
            unet_conversion_map_layer[sd_down_atn_prefix] = hf_down_atn_prefix

    for j in range(3):
        # loop over resnets/attentions for upblocks
        hf_up_res_prefix = f"up_blocks.{i}.resnets.{j}."
        sd_up_res_prefix = f"output_blocks.{3*i + j}.0."
        unet_conversion_map_layer[sd_up_res_prefix] = hf_up_res_prefix

        if i > 0:
            # no attention layers in up_blocks.0
            hf_up_atn_prefix = f"up_blocks.{i}.attentions.{j}."
            sd_up_atn_prefix = f"output_blocks.{3*i + j}.1."
            unet_conversion_map_layer[sd_up_atn_prefix] = hf_up_atn_prefix

    if i < 3:
        # no downsample in down_blocks.3
        hf_downsample_prefix = f"down_blocks.{i}.downsamplers.0.conv."
        sd_downsample_prefix = f"input_blocks.{3*(i+1)}.0.op."
        unet_conversion_map_layer[sd_downsample_prefix] = hf_downsample_prefix

        # no upsample in up_blocks.3
        hf_upsample_prefix = f"up_blocks.{i}.upsamplers.0."
        sd_upsample_prefix = f"output_blocks.{3*i + 2}.{1 if i == 0 else 2}."
        unet_conversion_map_layer[sd_upsample_prefix] = hf_upsample_prefix

hf_mid_atn_prefix = "mid_block.attentions.0."
sd_mid_atn_prefix = "middle_block.1."
unet_conversion_map_layer[sd_mid_atn_prefix] = hf_mid_atn_prefix

for j in range(2):
    hf_mid_res_prefix = f"mid_block.resnets.{j}."
    sd_mid_res_prefix = f"middle_block.{2*j}."
    unet_conversion_map_layer[sd_mid_res_prefix] = hf_mid_res_prefix


def convert_clip_l_key(ldm_key: str) -> str:
    kohya_key = '.'.join(ldm_key.split(".")[2:])  # cond_stage_model.transformer.
    kohya_key = f"te.{kohya_key}"
    return kohya_key
